"""Data matrix operators."""

from abc import abstractmethod
from typing import Generator, Self

import numpy
import pydantic

from tidyms2.core.dataflow import DataMatrixProcessStatus
from tidyms2.core.enums import MSInstrument, Polarity, SeparationMode

from .core.enums import AggregationMethod, CorrelationMethod, SampleType
from .core.matrix import DataMatrix
from .core.operators.matrix import ColumnFilter, MatrixTransformer, RowFilter
from .core.utils.numpy import FloatArray, FloatArray1D
from .core.utils.transformation import aggregate


class SampleFilter(RowFilter):
    """Remove samples using metadata properties.

    The :py:func:`~tidyms2.core.matrix.DataMatrix.Query.filter` query function is used to
    select which samples to remove. All samples matching the filter criteria
    will be removed from the data matrix.

    **Examples**

    Remove samples from `group-c` or `group-d`:

    .. code-block:: python

        from tidyms2.operators.SampleFilter

        op = SampleFilter(group=["group-c", "group-d"]})

    Remove all samples from the second analytical batch:

    .. code-block:: python

        from tidyms2.operators.SampleFilter

        op = SampleFilter(batch=2})

    Remove all blank samples:

    .. code-block:: python

        from tidyms2.operators.SampleFilter

        op = SampleFilter(type="blank")

    """

    model_config = pydantic.ConfigDict(extra="allow")

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()

    def _create_remove_list(self, data: DataMatrix) -> list[str]:
        filters = self.model_extra or dict()
        query = data.query.filter(**filters).fetch_sample_ids()
        return query[0][1] if query else list()


class FeatureFilter(ColumnFilter):
    """Remove feature groups based on descriptor values.

    This filter compares feature descriptors against lower and upper bounds and remove all features
    outside these values.

    **Examples**

    Remove all features with retention time lower than 30 s:

    .. code-block:: python

        from math import inf

        from tidyms2.operators import FeatureFilter

        op = FeatureFilter(rt=(30.0, inf))

    Remove all features with peak widths larger than 20 s:

    .. code-block:: python

        from tidyms2.operators import FeatureFilter

        op = FeatureFilter(width=(0.0, 20.0))

    """

    model_config = pydantic.ConfigDict(extra="allow")

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()

    @pydantic.model_validator(mode="after")
    def _ensure_tuple_in_filters(self):
        if self.model_extra is None:
            return self

        for field, value in self.model_extra.items():
            assert isinstance(value, tuple), "filter bounds must be a pair"
            assert len(value) == 2
            lb, ub = value
            assert lb <= ub, "The values in the filter should be a pair of lower and upper bounds."
        return self

    def _create_remove_list(self, data: DataMatrix) -> list[int]:
        filters = self.model_extra or dict()
        return [x.group for x in data.features if not x.has_descriptors_in_range(**filters)]


class BlankCorrector(MatrixTransformer):
    """Removes blank contribution from samples.

    The blank contribution is estimated as the aggregation of all samples that match the `blank_group`.
    If `blank_group` is not defined, all samples with blank type are used. The aggregation operation
    is defined by the `aggregation` parameter. By default, the blank contribution is estimated as the
    sample mean of all blank samples.

    Samples or features may be included from the correction by using the parameters `exclude_samples`
    and `exclude_features` respectively.

    """

    aggregation: AggregationMethod = AggregationMethod.MEAN
    """The method used to compute the blank contribution."""

    blank_groups: list[str] | None = None
    """The sample groups used to compute the blank contribution. If not specified, all samples
    from with BLANK :term:`sample type` will be used compute the blank contribution."""

    exclude_samples: list[str] | None = None
    """A list of sample ids to exclude from blank correction."""

    exclude_features: list[int] | None = None
    """A list of feature groups to exclude from blank correction."""

    apply_to_blanks: bool = False
    """Wether to apply the blank correction to samples used to compute the blank contribution."""

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()

    def _find_blank_sample_ids(self, data: DataMatrix) -> list[str]:
        if self.blank_groups is None:
            blank_samples_query = data.query.filter(type=SampleType.BLANK)
        else:
            blank_samples_query = data.query.filter(group=self.blank_groups)

        blank_sample_query_results = blank_samples_query.fetch_sample_ids()
        if not blank_sample_query_results:
            raise ValueError("No samples found in the blank group.")
        return blank_sample_query_results[0][1]

    def _transform_matrix(self, data: DataMatrix) -> FloatArray:
        blank_ids = self._find_blank_sample_ids(data)
        blank_data = data.get_data(sample_ids=blank_ids)
        blank_contribution = aggregate(blank_data, self.aggregation, axis=0)

        include_samples = list()
        exclude_samples = self.exclude_samples or list()
        for sample in data.samples:
            if sample.id in exclude_samples:
                continue
            if not self.apply_to_blanks and sample.id in blank_ids:
                continue
            include_samples.append(sample.id)
        include_samples_index = data.get_sample_index(*include_samples)

        include_features = list()
        exclude_features = self.exclude_features or list()
        for feature in data.features:
            if feature.group in exclude_features:
                continue
            include_features.append(feature.group)
        include_features_index = data.get_feature_index(*include_features)

        X = data.get_data().copy()
        for idx in include_samples_index:
            X[idx, include_features_index] -= blank_contribution[include_features_index]
        return numpy.maximum(X, 0.0)


class MetricsFilter(ColumnFilter):
    """Abstract base class for all metric-based filters."""

    lb: float = 0.0
    """The filter lower bound. Features with metric values lower than this value will be removed.
    Must be lower than `ub`."""

    ub: float = 0.25
    """The filter upper bound. Features with metric values greater than this value will be removed."""

    group_by: list[str] | None = None
    """How to group samples using sample metadata fields. If this parameter is specified, the feature
    metric will be computed for all groups and then aggregated using the `agg` parameter. For more details
    refer to the :py:func:`~tidyms2.core.matrix.DataMatrix.Query.group_by` method."""

    filter: dict | None = None
    """How to filter samples using sample metadata fields. This parameter is used to define which samples
    to include when computing the metric value. If not defined, no filtering is applied. For more details
    refer to the :py:func:`~tidyms2.core.matrix.DataMatrix.Query.filter` method."""

    aggregation: AggregationMethod = AggregationMethod.MIN
    """When the `group_by` parameter is defined, this parameter defines the aggregation method used to
    compute the feature value that will be compared against the filter lower and upper bounds. If
    `group_by` is not specified, this parameter is ignored."""

    @pydantic.model_validator(mode="after")
    def _ensure_lb_lower_than_ub(self) -> Self:
        if self.ub < self.lb:
            raise ValueError("Lower bound must be lower than the upper bound.")
        return self

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()

    @abstractmethod
    def _get_filter(self) -> dict | None: ...

    @abstractmethod
    def _get_group_by(self) -> list[str] | None: ...

    def _iter_groups(self, data: DataMatrix) -> Generator[DataMatrix, None, None]:
        filters = self._get_filter() or dict()
        groups = self._get_group_by() or tuple()
        for _, ids in data.query.filter(**filters).group_by(*groups).fetch_sample_ids():
            yield data.create_submatrix(sample_ids=ids)

    @abstractmethod
    def _compute_metric(self, data: DataMatrix) -> FloatArray1D: ...

    def aggregate_metric(self, data: DataMatrix) -> FloatArray:
        metric_list = list()
        for matrix in self._iter_groups(data):
            metric_list.append(self._compute_metric(matrix))

        if not metric_list:
            raise ValueError("No sample groups found.")

        metric_array = numpy.vstack(metric_list)
        return aggregate(metric_array, self.aggregation)

    def _create_remove_list(self, data: DataMatrix) -> list[int]:
        metric = self.aggregate_metric(data)
        ind = numpy.where((metric < self.lb) | (metric > self.ub))[0]
        return [data.features[x.item()].group for x in ind]


class CVFilter(MetricsFilter):
    """Filter features using the coefficient of variation (CV) metric.

    This filter allows to compute the CV on multiple sample subsets by using the matrix
    :py:class:`~tidyms2.core.matrix.DataMatrix.Query` utilities. The parameters `group_by`
    and `filter` are used to this end. Refer to the :ref:`matrix-query-guide` guide for
    more details.

    If neither the `group_by` and `filter` parameters are defined, the CV is computed on
    all samples labelled with the QC :term:`sample type`. If QC samples are found, a
    ``ValuerError`` is raised.

    If multiple sample subsets are used, an aggregated CV value is obtained
    for each feature by using the `aggregation` parameter, which defines the aggregation
    operation. By default, the minimum CV of all groups is kept. The aggregated CV
    value is then compared against the filter lower and upper bounds, defined by the
    `lb` and `ub` parameters. Features with values outside these bounds are removed from
    the data matrix.

    The :ref:`data-curation-guide` guide covers several examples of this filter.

    .. seealso::
        :py:func:`~tidyms2.core.matrix.DataMatrix.Metrics.cv`: The CV metric computation
        method.

    """

    robust: bool = True
    """Compute the CV using a robust estimator."""

    def _get_filter(self) -> dict | None:
        if self.filter is None and self.group_by is None:
            return {"type": SampleType.TECHNICAL_QC}
        return self.filter

    def _get_group_by(self) -> list[str] | None:
        return self.group_by

    def _compute_metric(self, data: DataMatrix) -> FloatArray1D:
        return data.metrics.cv(robust=self.robust)


class DRatioFilter(MetricsFilter):
    """Filter features using the D-ratio metric.

    D-ratio is computed for each feature and features are removed if they fall outside the
    filters lower and upper bounds. By default, the sample variance is estimated using all
    samples with sample type. The technical variation is estimated all QC samples.

    The :ref:`data-curation-guide` guide covers several examples of this filter.

    .. seealso::
        :py:func:`~tidyms2.core.matrix.DataMatrix.Metrics.dratio`: The D-Ratio metric computation
        method.

    """

    lb: pydantic.NonNegativeFloat = 0.0
    ub: pydantic.NonNegativeFloat = 1.0

    robust: bool = True
    """Compute the D-ratio using a robust estimator."""

    sample_groups: list[str] | None = None
    """A list of :term:`sample group` used to estimate the biological variation. If not defined, all samples
    with sample type are used."""

    qc_groups: list[str] | None = None
    """A list of :term:`sample group` used to estimate the technical variation. If not defined, all samples
    with QC type are used."""

    def _get_filter(self):
        return None

    def _get_group_by(self):
        return None

    @pydantic.field_validator("filter", "group_by", mode="before")
    @classmethod
    def _ensure_none(cls, val, info: pydantic.ValidationInfo):
        if val is not None:
            raise ValueError(f"{info.field_name} usage is not supported for this filter.")
        return val

    def _compute_metric(self, data: DataMatrix) -> FloatArray1D:
        return data.metrics.dratio(self.sample_groups, self.qc_groups, robust=self.robust)


class DetectionRateFilter(MetricsFilter):
    """Filter features using the detection rate (DR) metric.

    This filter allows to compute the DR on multiple sample subsets by using the matrix
    :py:class:`~tidyms2.core.matrix.DataMatrix.Query` utilities. The parameters `group_by`
    and `filter` are used to this end. Refer to the :ref:`matrix-query-guide` guide for
    more details.

    If neither the `group_by` and `filter` parameters are defined, samples are grouped
    by :term:`sample group` and filtered by :term:`sample type`, keeping only samples
    with sample type, before computing the DR on each group. If no samples with this type
    are found, a ``ValuerError`` is raised.

    By default, the maximum DR value of all groups is kept. The aggregated DR
    value is then compared against the filter lower and upper bounds, defined by the
    `lb` and `ub` parameters. Features with values outside these bounds are removed from
    the data matrix.

    The :ref:`data-curation-guide` guide covers the usage of this filter.

    .. seealso::
        :py:func:`~tidyms2.core.matrix.DataMatrix.Metrics.detection_rate`: The detection
        rate metric computation method.

    """

    lb: pydantic.NonNegativeFloat = 0.9
    ub: pydantic.NonNegativeFloat = 1.0

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    threshold: float | FloatArray1D | None = None
    """The threshold value to define a feature as detected. If a scalar is provided, then
    the same threshold is applied to all features. Using an array allows to use an unique
    threshold for each feature. If the threshold is set to ``None``, the threshold value
    will be estimated using :py:func:`~tidyms2.core.matrix.DataMatrix.Metrics.lod`."""

    def _get_filter(self) -> dict | None:
        if self.group_by is None and self.filter is None:
            return {"type": SampleType.SAMPLE}
        return self.filter

    def _get_group_by(self) -> list[str] | None:
        if self.group_by is None and self.filter is None:
            return ["group"]
        return self.group_by

    def _compute_metric(self, data: DataMatrix) -> FloatArray1D:
        threshold = self.threshold or data.metrics.lod()
        return data.metrics.detection_rate(threshold=threshold)


class CorrelationFilter(MetricsFilter):
    """Filter features based on the correlation with sample metadata fields.

     This filter allows to compute the correlation with a sample metadata field on multiple
     sample subsets by using the matrix :py:class:`~tidyms2.core.matrix.DataMatrix.Query`
     utilities. The parameters `group_by` and `filter` are used to this end. Refer to the
     :ref:`matrix-query-guide` guide for more details.

    If neither the `group_by` and `filter` parameters are defined, the correlation is
    computed using all samples.

    This filter will fail if the data contain NaN values.

    **Examples**

    Filter features using the correlation between the run order and intensity on QC samples:

    .. code-block:: python

        from tidyms2.operators import CorrelationFilter

        op = CorrelationFilter(filter={"type": "qc"}, field="order")

    Filter features using the correlation between the dilution factor and intensity of diluter QC samples:

    .. code-block:: python

        from tidyms2.operators import CorrelationFilter

        op = CorrelationFilter(filter={"type": "dqc"}, field="dilution")

    """

    field: str
    """The sample metadata field used to compute the correlation."""

    method: CorrelationMethod = CorrelationMethod.PEARSON
    """The method used to compute the correlation."""

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus(missing_imputed=True)

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus(missing_imputed=True)

    def _get_filter(self):
        return None

    def _get_group_by(self):
        return None

    def _compute_metric(self, data: DataMatrix) -> FloatArray1D:
        return data.metrics.correlation(self.field, self.method)



"""Transformation utilities for numpy arrays."""

from typing import assert_never

import numpy

from ..enums import AggregationMethod, NormalizationMethod, ScalingMethod
from .numpy import FloatArray, FloatArray1D, check_matrix_shape


def aggregate(X: FloatArray, method: AggregationMethod | str, axis: int = 0) -> FloatArray1D:
    """Apply an aggregation operation to an array."""
    check_matrix_shape(X)

    if not isinstance(method, AggregationMethod):
        method = AggregationMethod(method)

    match method:
        case AggregationMethod.LOD:
            return numpy.mean(X, axis=axis) + 3 * numpy.std(X, axis=axis, ddof=1)
        case AggregationMethod.LOQ:
            return numpy.mean(X, axis=axis) + 10 * numpy.std(X, axis=axis, ddof=1)
        case AggregationMethod.MIN:
            return numpy.min(X, axis=axis)
        case AggregationMethod.MEAN:
            return numpy.mean(X, axis=axis)
        case AggregationMethod.MEDIAN:
            return numpy.median(X, axis=axis)
        case AggregationMethod.MAX:
            return numpy.max(X, axis=axis)
        case AggregationMethod.SUM:
            return numpy.sum(X, axis=axis)
        case AggregationMethod.STD:
            return numpy.std(X, axis=axis, ddof=1)
        case _ as never:
            assert_never(never)


def scale(X: FloatArray, method: ScalingMethod | str = ScalingMethod.AUTOSCALING) -> FloatArray:
    """Scale the columns of a 2D array.

    :param X: The 2D array to scale
    :method: Available scaling methods.``"autoscaling"`` applies mean centering and scaling to
        unitary population variance. ``"rescaling"`` scales data to the :math:`[0, 1]` range. ``"pareto"``
        performs mean centering and scaling using the square root of the population standard deviation
    :return: A scaled 2D array

    """
    check_matrix_shape(X)

    if not isinstance(method, ScalingMethod):
        method = ScalingMethod(method)

    match method:
        case ScalingMethod.AUTOSCALING:
            return (X - numpy.nanmean(X, axis=0, keepdims=True)) / numpy.nanstd(X, axis=0, keepdims=True)
        case ScalingMethod.RESCALING:
            Xmin = numpy.nanmin(X, axis=0, keepdims=True)
            return (X - Xmin) / (numpy.nanmax(X, axis=0, keepdims=True) - Xmin)
        case ScalingMethod.PARETO:
            return (X - numpy.nanmean(X, axis=0, keepdims=True)) / numpy.sqrt(numpy.nanstd(X, axis=0, keepdims=True))
        case _ as never:
            assert_never(never)


def normalize(
    X: FloatArray, method: NormalizationMethod | str = NormalizationMethod.SUM, index: int | None = None
) -> FloatArray:
    """Normalize rows of a 2D array.

    :param X: the 2D array to normalize
    :param method: the normalization method. ``"sum"`` normalizes rows using the sum of values in each row,
        ``"max"` normalizes rows using the maximum value of each row. ``"euclidean"`` normalizes rows
        using the euclidean norm of the row. ``"feature"`` normalizes rows using the value of a column.
    :param index: the column index used for ``"feature"`` normalization. This parameter for all other feature
        normalization methods.
    :return: the normalized array. Rows values may be set to NaN or Inf in the following cases: for `"sum"`,
        `"max"` or `"euclidean"` normalization, if all values in the row are `0.0` or NaN, all values will be
        set to `NaN`. For `"feature"` NaN value will be obtained if the reference column value is NaN or if
        column value is NaN. NaN values will also be obtained if the reference feature and column value are both
        0.0. If only the reference feature value is 0.0, then Inf will be obtained.

    """
    check_matrix_shape(X)

    if not isinstance(method, NormalizationMethod):
        method = NormalizationMethod(method)

    if method is NormalizationMethod.FEATURE and index is None:
        raise ValueError("`index` must be specified for feature normalization.")

    match method:
        case NormalizationMethod.SUM:
            factor = numpy.nansum(X, axis=1, keepdims=True)
        case NormalizationMethod.MAX:
            factor = numpy.nanmax(X, axis=1, keepdims=True)
        case NormalizationMethod.EUCLIDEAN:
            factor = numpy.sqrt(numpy.nansum(X**2, axis=1, keepdims=True))
        case NormalizationMethod.FEATURE:
            factor = X[:, index]
        case _ as never:
            assert_never(never)
    return X / factor



"""Data models for working with LC-MS data."""

from __future__ import annotations

import bisect
from functools import cached_property
from math import nan

import numpy as np
import pydantic
from scipy.integrate import cumulative_trapezoid, trapezoid
from typing_extensions import Self

from ..core.models import AnnotableFeature, IsotopicEnvelope, MZTrace


class Peak(AnnotableFeature[MZTrace]):
    """Representation of a chromatographic peak."""

    start: pydantic.NonNegativeInt
    """index in the m/z trace where the peak begins. Must be smaller than `apex`"""

    apex: pydantic.PositiveInt
    """index in the m/z trace where the apex of the peak is located. Must be smaller than `end`"""

    end: pydantic.PositiveInt
    """index in the m/z trace where the peak ends. Start and end used as slices defines the peak region."""

    @pydantic.model_validator(mode="after")
    def check_peak_definition(self) -> Self:
        """Perform peak sanity check.

        Check that start, apex and end indices are strictly sorted. Check that
        peak end index is lower than trace length.

        """
        msg = "start must be lower than loc and loc must be lower than end"
        assert self.start < self.apex < self.end, msg

        msg = "peak end must be lower than m/z trace length."
        assert self.end <= self.roi.scan.size, msg

        return self

    @pydantic.computed_field(repr=False)
    @cached_property
    def rt_start(self) -> float:
        """The peak start time."""
        return self.roi.time[self.start].item()

    @pydantic.computed_field(repr=False)
    @cached_property
    def rt_end(self) -> float:
        """The peak end time."""
        return self.roi.time[self.end - 1].item()

    @pydantic.computed_field
    @cached_property
    def rt(self) -> float:
        """Peak retention time, defined as the weighted average of the trace time in the peak region.

        The trace height is used as weights.
        """
        try:
            weights = self.roi.get_slice_height(self.start, self.end)
            return np.average(self.roi.time[self.start : self.end], weights=weights).item()
        except ZeroDivisionError:
            return nan

    @pydantic.computed_field(repr=False)
    @cached_property
    def height(self) -> float:
        """Peak height, defined as the difference between the peak intensity and the peak baseline at the apex."""
        height = self.roi.spint[self.apex].item()
        if self.roi.baseline is not None:
            height = height - self.roi.baseline[self.apex].item()
        return height

    @pydantic.computed_field(repr=False)
    @cached_property
    def area(self) -> float:
        """The peak area."""
        peak_extension = self.roi.get_slice_height(self.start, self.end)
        return trapezoid(peak_extension, self.roi.time[self.start : self.end]).item()

    @pydantic.computed_field(repr=False)
    @cached_property
    def width(self) -> float:
        """Compute the peak width.

        The peak width is defined as the region where the 95 % of the total peak
        area is distributed.

        Returns
        -------
        width : positive number.

        """
        height = self.roi.get_slice_height(self.start, self.end)

        area = cumulative_trapezoid(height, self.roi.time[self.start : self.end])
        if area[-1] > 0:
            relative_area = area / area[-1]
            percentile = [0.025, 0.975]
            start, end = self.start + np.searchsorted(relative_area, percentile)
            return float(self.roi.time[end] - self.roi.time[start])
        else:
            return nan

    @pydantic.computed_field(repr=False)
    @cached_property
    def extension(self) -> float:
        """The peak extension, defined as the length of the peak region."""
        return (self.roi.time[self.end - 1] - self.roi.time[self.start]).item()

    @pydantic.computed_field(repr=False)
    @cached_property
    def snr(self) -> float:
        """The peak signal-to-noise ratio.

        The SNR is defined as the quotient between the peak height and the noise level at the apex.
        If the noise level is not available, the SNR is set to ``nan``.
        """
        if self.roi.noise is None or np.allclose(self.roi.noise[self.apex], 0.0):
            return nan
        else:
            return (self.get("height") / self.roi.noise[self.apex]).item()

    @pydantic.computed_field
    @cached_property
    def mz(self) -> float:
        """The peak m/z, defined as the weighted average of the trace m/z in the peak region.

        The trace height is used as weights.
        """
        weights = self.roi.get_slice_height(self.start, self.end)
        try:
            mz_mean = np.average(self.roi.mz[self.start : self.end], weights=weights)
            return max(0.0, mz_mean.item())
        except ZeroDivisionError:
            return nan

    @pydantic.computed_field(repr=False)
    @cached_property
    def mz_std(self) -> float:
        """The peak m/z standard deviation."""
        return self.roi.mz[self.start : self.end].std().item()

    def compare(self, other: Self) -> float:
        """Compute the similarity between a pair of peaks.

        The similarity is defined as the cosine distance between the overlapping
        region of two peaks.

        """
        return _compare_features_lc(self, other)

    @staticmethod
    def compute_isotopic_envelope(*features) -> IsotopicEnvelope:
        """Compute the isotopic envelope (m/z and abundance) of a list of peaks.

        :param features: the peaks that conform the envelope

        """
        scan_start = 0
        scan_end = 10000000000  # dummy value
        for ft in features:
            scan_start = max(scan_start, ft.roi.scan[ft.start])
            scan_end = min(scan_end, ft.roi.scan[ft.end - 1])

        p = list()
        mz = list()
        if scan_start < scan_end:
            for ft in features:
                start = bisect.bisect(ft.roi.scan, scan_start)
                end = bisect.bisect(ft.roi.scan, scan_end)
                apex = (start + end) // 2  # dummy value
                tmp_peak = Peak(start=start, apex=apex, end=end, roi=ft.roi)
                p_area = trapezoid(tmp_peak.roi.spint[start:end], tmp_peak.roi.time[start:end])
                p.append(p_area)
                mz.append(ft.mz)
        total_area = sum(p)
        p = [x / total_area for x in p]
        return IsotopicEnvelope(mz=mz, p=p)


def _compare_features_lc(ft1: Peak, ft2: Peak) -> float:
    """Feature similarity function used in LC-MS data."""
    start1 = ft1.roi.scan[ft1.start]
    start2 = ft2.roi.scan[ft2.start]
    if start1 > start2:
        ft1, ft2 = ft2, ft1
    overlap_ratio = _overlap_ratio(ft1, ft2)
    min_overlap = 0.5
    if overlap_ratio > min_overlap:
        os1, oe1, os2, oe2 = _get_overlap_index(ft1, ft2)
        norm1 = np.linalg.norm(ft1.roi.spint[ft1.start : ft1.end])
        norm2 = np.linalg.norm(ft2.roi.spint[ft2.start : ft2.end])
        x1 = ft1.roi.spint[os1:oe1] / norm1
        x2 = ft2.roi.spint[os2:oe2] / norm2
        similarity = np.dot(x1, x2)
    else:
        similarity = 0.0
    return similarity


def _overlap_ratio(ft1: Peak, ft2: Peak) -> float:
    """Compute the overlap ratio, between a pair of peaks.

    The overlap ratio is the quotient between the overlap region and the
    maximum value of the extension.

    """
    start2 = ft2.roi.scan[ft2.start]
    end1 = ft1.roi.scan[ft1.end - 1]
    end2 = ft2.roi.scan[ft2.end - 1]
    # start1 <= start2. end1 > start2 is a sufficient condition for overlap
    if end1 > start2:
        # the overlap ratio is the quotient between the length overlapped region
        # and the extension of the shortest feature.
        if end1 <= end2:
            start2_index_in1 = bisect.bisect_left(ft1.roi.scan, start2)
            overlap_length = ft1.end - start2_index_in1
        else:
            overlap_length = ft2.end - ft2.start
        min_length = min(ft1.end - ft1.start, ft2.end - ft2.start)
        res = overlap_length / min_length
    else:
        res = 0.0
    return res


def _get_overlap_index(ft1: Peak, ft2: Peak) -> tuple[int, int, int, int]:
    """Compute the overlap indices for ft1 and ft2.

    `ft1` must start before `ft2`

    """
    end1 = ft1.roi.scan[ft1.end - 1]
    end2 = ft2.roi.scan[ft2.end - 1]
    start2 = ft2.roi.scan[ft2.start]
    if end1 >= end2:
        overlap_start1 = bisect.bisect_left(ft1.roi.scan, start2)
        overlap_end1 = bisect.bisect(ft1.roi.scan, end2)
        overlap_start2 = ft2.start
        overlap_end2 = ft2.end
    else:
        overlap_start1 = bisect.bisect_left(ft1.roi.scan, start2)
        overlap_end1 = ft1.end
        overlap_start2 = ft2.start
        overlap_end2 = bisect.bisect(ft2.roi.scan, end1)
    return overlap_start1, overlap_end1, overlap_start2, overlap_end2



"""Annotate function implementation."""

from typing import Sequence

from ..chem import DEFAULT_CONTEXT, EM, ChemicalContext, EnvelopeValidator
from ..core.models import AnnotableFeature, AnnotableFeatureType
from .annotation_data import AnnotationData
from .config import AnnotatorParameters
from .envelope_finder import EnvelopeFinder
from .mmi_finder import MMIFinder


def create_annotation_tools(
    config: AnnotatorParameters, context: ChemicalContext | None = None
) -> tuple[MMIFinder, EnvelopeFinder, EnvelopeValidator]:
    """Create an annotator object. Auxiliary function to _annotate.

    :param config: the annotation parameters
    :param context: the chemical context used in during annotation

    """
    context = DEFAULT_CONTEXT if context is None else context
    bin_size = 100
    mmi_finder: MMIFinder = MMIFinder(config, bin_size, context)
    envelope_finder: EnvelopeFinder = EnvelopeFinder(config, context)
    envelope_validator = EnvelopeValidator(config)
    return mmi_finder, envelope_finder, envelope_validator


def annotate(
    feature_list: Sequence[AnnotableFeature],
    mmi_finder: MMIFinder,
    envelope_finder: EnvelopeFinder,
    envelope_validator: EnvelopeValidator,
) -> None:
    """Annotate isotopologues in a sample.

    Annotations are added to the `annotation` attribute of each feature.

    :param feature_list: list of features obtained after feature extraction.
    :param mmi_finder: MMIFinder
    :param envelope_finder: EnvelopeFinder
    :param envelope_validator: EnvelopeValidator

    """
    data = AnnotationData(feature_list)
    monoisotopologue = data.get_monoisotopologue()
    polarity = mmi_finder.polarity
    while monoisotopologue is not None:
        mmi_candidates = mmi_finder.find(data)
        envelope, charge = find_best_envelope(
            data,
            monoisotopologue,
            polarity,
            mmi_candidates,
            envelope_finder,
            envelope_validator,
        )
        data.annotate(envelope, charge)
        monoisotopologue = data.get_monoisotopologue()


def find_best_envelope(
    data: AnnotationData[AnnotableFeatureType],
    monoisotopologue: AnnotableFeatureType,
    polarity: int,
    mmi_candidates: Sequence[tuple[AnnotableFeatureType, int]],
    envelope_finder: EnvelopeFinder,
    envelope_validator: EnvelopeValidator,
) -> tuple[Sequence[AnnotableFeatureType], int]:
    """Find the most fitting envelope candidate."""
    best_length = 1
    best_candidate = [monoisotopologue]
    best_charge = -1
    for mmi, charge in mmi_candidates:
        envelope_candidates = envelope_finder.find(data, mmi, charge)
        for candidate in envelope_candidates:
            validated_length = _validate_candidate(
                candidate,
                monoisotopologue,
                charge,
                polarity,
                best_length,
                envelope_validator,
            )
            if validated_length > best_length:
                best_length = validated_length
                best_candidate = candidate[:validated_length]
                best_charge = charge
    return best_candidate, best_charge


def _validate_candidate(
    candidate: Sequence[AnnotableFeatureType],
    monoisotopologue: AnnotableFeatureType,
    charge: int,
    polarity: int,
    min_length: int,
    validator: EnvelopeValidator,
) -> int:
    if len(candidate) <= min_length:
        return 0

    if monoisotopologue not in candidate:
        return 0

    envelope = candidate[0].compute_isotopic_envelope(*candidate)
    em_correction = EM * charge * polarity
    M = [x * charge - em_correction for x in envelope.mz]
    return validator.validate(M, envelope.p)



import pytest

from tidyms2.annotation import annotation
from tidyms2.core.models import MZTrace
from tidyms2.core.operators.pipeline import Pipeline
from tidyms2.lcms import LCPeakExtractor, LCTraceBaselineEstimator, LCTraceExtractor, Peak
from tidyms2.simulation.lcms import SimulatedLCMSSampleFactory
from tidyms2.storage.memory import OnMemorySampleStorage


@pytest.fixture
def annotation_tools_params():
    return annotation.AnnotatorParameters(
        bounds={
            "C": (0, 50),
            "H": (0, 100),
            "O": (0, 20),
            "N": (0, 20),
            "Cl": (0, 2),
            "B": (0, 1),
        },
        max_M=2500,
        max_length=10,
        max_charge=3,
        min_M_tol=0.005,
        max_M_tol=0.01,
        p_tol=0.05,
        min_similarity=0.9,
        min_p=0.01,
    )


def test__annotate_empty_feature_list(annotation_tools_params):
    tools = annotation.create_annotation_tools(annotation_tools_params)
    feature_list = list()
    annotation.annotate(feature_list, *tools)


@pytest.fixture
def peak_list() -> list[Peak]:
    factory_config = {
        "config": {"n_scans": 300, "amp_noise": 0.0},
        "adducts": [
            {
                "formula": "[C10H20O2]-",
                "abundance": {"mu": 10000},
                "rt_mean": 50,
                "rt_width": 3.0,
                "n_isotopologues": 4,
            },
            {
                "formula": "[C10H20BO3]-",
                "abundance": {"mu": 20000},
                "rt_mean": 75,
                "rt_width": 3.0,
                "n_isotopologues": 4,
            },
            {
                "formula": "[C20H40BO5]2-",
                "abundance": {"mu": 30000},
                "rt_mean": 150,
                "rt_width": 3.0,
                "n_isotopologues": 4,
            },
            {
                "formula": "[C18H19N2O3]-",
                "abundance": {"mu": 25000},
                "rt_mean": 200,
                "rt_width": 3.0,
                "n_isotopologues": 4,
            },
            {
                "formula": "[C18H20N2O3Cl]-",
                "abundance": {"mu": 25000},
                "rt_mean": 200,
                "rt_width": 3.0,
                "n_isotopologues": 4,
            },
            {
                "formula": "[C10H20Cl]-",
                "abundance": {"mu": 20000},
                "rt_mean": 175,
                "rt_width": 3.0,
                "n_isotopologues": 4,
            },
        ],
    }
    factory = SimulatedLCMSSampleFactory(**factory_config)
    sample = factory("simulated_sample")

    data = OnMemorySampleStorage(sample, MZTrace, Peak)

    pipe = Pipeline(id="annotation_pipeline")
    pipe.add_operator(LCTraceExtractor(id="roi_extractor"))
    pipe.add_operator(LCTraceBaselineEstimator(id="baseline_estimator"))
    pipe.add_operator(LCPeakExtractor(id="peak_extractor"))
    pipe.apply(data)

    return data.list_features()


def test_annotate(peak_list: list[Peak], annotation_tools_params):
    tools = annotation.create_annotation_tools(annotation_tools_params)
    annotation.annotate(peak_list, *tools)

    # group features by isotopologue label.
    annotation_check = dict()
    for peak in peak_list:
        assert peak.annotation is not None
        group_list = annotation_check.setdefault(peak.annotation.isotopologue_label, list())
        group_list.append(peak)
    if -1 in annotation_check:
        annotation_check.pop(-1)
    assert len(annotation_check) == 6
    for v in annotation_check.values():
        assert len(v) == 4  # features where generated with 4 isotopologues.



"""Utilities to read matrix data in a variety of formats."""

import csv
from pathlib import Path

from numpy import array

from ..core.matrix import DataMatrix
from ..core.models import FeatureGroup, GroupAnnotation, Sample, SampleMetadata

PROGENESIS_WIDTH = "Chromatographic peak width (min)"
PROGENESIS_RT = "Retention time (min)"
PROGENESIS_CHARGE = "Charge"
PROGENESIS_MZ = "m/z"
PROGENESIS_NAME = "Compound"
NORMALIZED_HEADER_FIELD = "Normalised abundance"
RAW_HEADER_FIELD = "Raw abundance"


class ProgenesisReaderError(ValueError):
    """Exception raised when parsing a progenesis file fails."""


def read_progenesis(path: Path) -> DataMatrix:
    """Read progenesis CSV data into a data matrix.

    :param path: Path to the Progenesis data file.

    """
    with path.open() as fp:
        reader = csv.reader(fp)

        # first row contains the location of normalized and raw data
        first_row = next(reader)
        ft_meta_start, ft_meta_end, data_start, data_end = _get_data_location(first_row)

        # second row contain sample groups
        second_row = next(reader)
        sample_groups = _get_sample_groups(second_row, data_start, data_end)

        # third row contain sample ids and feature metadata fields
        third_row = next(reader)

        sample_ids = third_row[data_start:data_end]
        ft_meta_fields = third_row[ft_meta_start:ft_meta_end]

        samples = _create_samples(sample_ids, sample_groups)

        meta_field_to_index = _map_ft_meta_field_to_index(ft_meta_fields, ft_meta_start)

        columns = list()
        feature_groups = list()
        for k, row in enumerate(reader):
            feature_groups.append(_create_feature_group(row, meta_field_to_index, k))
            columns.append(_create_matrix_column(row, data_start, data_end))

        X = array(columns).T

    return DataMatrix(samples, feature_groups, X)


def _get_data_location(first_row: list[str]) -> tuple[int, int, int, int]:
    """Find indices where feature metadata ends and matrix data starts."""
    try:
        ft_meta_end = first_row.index(NORMALIZED_HEADER_FIELD)
    except ValueError:
        raise ProgenesisReaderError(f"Header row does not contain header field `{NORMALIZED_HEADER_FIELD}`.")

    try:
        data_start = first_row.index(RAW_HEADER_FIELD)
    except ValueError:
        raise ProgenesisReaderError(f"Header row does not contain header field `{RAW_HEADER_FIELD}`.")

    ft_meta_start = 0
    data_start = data_start
    data_end = 2 * data_start - ft_meta_end
    return ft_meta_start, ft_meta_end, data_start, data_end


def _get_sample_groups(second_row: list[str], data_start: int, data_end: int) -> list[str]:
    groups = list()

    if not second_row:
        raise ProgenesisReaderError("Expected second data row to contain sample group data but found None.")

    prev = ""
    for g in second_row[data_start:data_end]:
        current = g or prev
        groups.append(current)
        prev = g
    return groups


def _create_samples(sample_ids: list[str], sample_groups: list[str]) -> list[Sample]:
    if len(sample_ids) != len(sample_groups):
        raise ProgenesisReaderError("Number of samples ids and sample groups in data do not match.")
    samples = list()
    for order, (id_, group) in enumerate(zip(sample_ids, sample_groups)):
        s = Sample(id=id_, path=Path("."), meta=SampleMetadata(order=order, group=group))
        samples.append(s)
    return samples


def _map_ft_meta_field_to_index(fields: list[str], ft_meta_start: int) -> dict[str, int]:
    """Create a mapping from field names to indices for efficient access to feature metadata."""
    field_to_index = dict()
    for field in [PROGENESIS_NAME, PROGENESIS_MZ, PROGENESIS_CHARGE, PROGENESIS_RT, PROGENESIS_WIDTH]:
        try:
            field_to_index[field] = fields.index(field) + ft_meta_start
        except ValueError:
            raise ProgenesisReaderError(f"`{field}` not found in data file.")
    return field_to_index


def _create_feature_group(row: list[str], field_to_index: dict[str, int], group: int) -> FeatureGroup:
    charge = int(row[field_to_index[PROGENESIS_CHARGE]])
    name = row[field_to_index[PROGENESIS_NAME]]  # progenesis internal name
    ann = GroupAnnotation(label=group, charge=charge, name=name)

    descriptors = {
        "mz": float(row[field_to_index[PROGENESIS_MZ]]),
        "rt": float(row[field_to_index[PROGENESIS_RT]]) * 60,  # progenesis express time in min
        "width": float(row[field_to_index[PROGENESIS_WIDTH]]) * 60,
    }
    return FeatureGroup(group=group, annotation=ann, descriptors=descriptors)


def _create_matrix_column(row: list[str], start: int, end: int) -> list[float]:
    return [float(x or 0.0) for x in row[start:end]]



import pydantic
import pytest

from tidyms2.core.dataflow import SampleProcessStatus
from tidyms2.core.exceptions import PipelineConfigurationError, ProcessStatusError, RepeatedIdError
from tidyms2.core.operators.pipeline import Pipeline

from ... import helpers


class TestRoiExtractor:
    @pytest.fixture
    def op(self):
        return helpers.DummyRoiExtractor()

    def test_set_invalid_parameter_raise_ValidationError(self, op):
        with pytest.raises(pydantic.ValidationError):
            op.param2 = 10  # type: ignore

    def test_update_process_status(self, op: helpers.DummyRoiExtractor):
        status_in = SampleProcessStatus()
        assert not status_in.roi_extracted
        op.update_status(status_in)
        assert status_in.roi_extracted

    def test_apply(self, sample_storage, op):
        assert not sample_storage.list_rois()
        op.apply(sample_storage)

        assert len(sample_storage.list_rois()) == op.n_roi
        assert sample_storage.get_status().roi_extracted


class TestRoiTransformer:
    @pytest.fixture
    def op(self):
        return helpers.DummyRoiTransformer()

    def test_check_status_invalid_status_raises_error(self, op: helpers.DummyRoiTransformer):
        status = SampleProcessStatus()
        with pytest.raises(ProcessStatusError):
            op.check_status(status)

    def test_check_status_ok(self, op: helpers.DummyRoiTransformer):
        status = SampleProcessStatus(roi_extracted=True)
        op.check_status(status)

    def test_apply(self, sample_storage_with_rois, op):
        op.apply(sample_storage_with_rois)

        assert all(x.data == op.max_length for x in sample_storage_with_rois.list_rois())
        assert sample_storage_with_rois.get_status().roi_extracted


class TestFeatureExtractor:
    @pytest.fixture
    def op(self):
        return helpers.DummyFeatureExtractor()

    def test_check_status_invalid_status_raises_error(self, op: helpers.DummyFeatureExtractor):
        status = SampleProcessStatus()
        with pytest.raises(ProcessStatusError):
            op.check_status(status)

    def test_check_status_ok(self, op: helpers.DummyFeatureExtractor):
        status = SampleProcessStatus(roi_extracted=True)
        op.check_status(status)

    def test_apply(self, sample_storage_with_rois, op):
        op.apply(sample_storage_with_rois)

        rois = sample_storage_with_rois.list_rois()
        assert rois
        for roi in rois:
            features = sample_storage_with_rois.list_features(roi_id=roi.id)
            assert len(features) == op.n_features

        assert sample_storage_with_rois.get_status().roi_extracted
        assert sample_storage_with_rois.get_status().feature_extracted

    def test_apply_with_filter_remove_all_features(self, sample_storage_with_rois, op):
        op.bounds = {"height": (1000000.0, 2000000.0)}
        op.apply(sample_storage_with_rois)

        assert sample_storage_with_rois.list_rois()
        assert not sample_storage_with_rois.list_features()

        assert sample_storage_with_rois.get_status().roi_extracted
        assert sample_storage_with_rois.get_status().feature_extracted

    def test_apply_with_filter_fill_lower_bound_with_inf(self, sample_storage_with_rois, op):
        op.bounds = {"height": (None, 2000000.0)}
        op.apply(sample_storage_with_rois)

        assert sample_storage_with_rois.list_rois()
        assert sample_storage_with_rois.list_features()

        assert sample_storage_with_rois.get_status().roi_extracted
        assert sample_storage_with_rois.get_status().feature_extracted

    def test_apply_with_filter_fill_upper_bound_with_inf(self, sample_storage_with_rois, op):
        op.bounds = {"height": (0.0, None)}
        op.apply(sample_storage_with_rois)

        assert sample_storage_with_rois.list_rois()
        assert sample_storage_with_rois.list_features()

        assert sample_storage_with_rois.get_status().roi_extracted
        assert sample_storage_with_rois.get_status().feature_extracted


class TestFeatureTransformer:
    @pytest.fixture
    def op(self):
        return helpers.DummyFeatureTransformer()

    def test_check_status_invalid_status_raises_error(self, op: helpers.DummyFeatureTransformer):
        status = SampleProcessStatus(roi_extracted=True)
        with pytest.raises(ProcessStatusError):
            op.check_status(status)

    def test_check_status_ok(self, op: helpers.DummyRoiTransformer):
        status = SampleProcessStatus(roi_extracted=True, feature_extracted=True)
        op.check_status(status)

    def test_apply(self, sample_storage_with_features, op):
        op.apply(sample_storage_with_features)

        assert all(x.data_mz == op.feature_value for x in sample_storage_with_features.list_features())
        assert sample_storage_with_features.get_status().roi_extracted


class TestPipeline:
    @pytest.fixture
    def pipe(self):
        return Pipeline(id="test-pipeline")

    def test_add_sample_operator_to_empty_pipeline(self, pipe):
        pipe.add_operator(helpers.DummyRoiExtractor())

    def test_add_assay_operator_to_empty_pipeline(self, pipe):
        pipe.add_operator(helpers.DummyAnnotationPatcher())

    def test_add_sample_operator_with_repeated_raises_error(self, pipe):
        op = helpers.DummyRoiExtractor(id="dummy-id")
        pipe.add_operator(op)
        with pytest.raises(RepeatedIdError):
            pipe.add_operator(op)

    def test_add_sample_operator_to_pipeline_with_assay_operator_raises_error(self, pipe):
        assay_op = helpers.DummyDescriptorPatcher(id="dummy-descriptor-patcher-id")
        pipe.add_operator(assay_op)
        with pytest.raises(PipelineConfigurationError):
            sample_op = helpers.DummyRoiExtractor(id="dummy-roi-extractor-id")
            pipe.add_operator(sample_op)

    def test_add_assay_operator_to_pipeline_with_sample_operator_raises_error(self, pipe):
        sample_op = helpers.DummyRoiExtractor(id="dummy-roi-extractor-id")
        pipe.add_operator(sample_op)
        with pytest.raises(PipelineConfigurationError):
            assay_op = helpers.DummyDescriptorPatcher(id="dummy-descriptor-patcher-id")
            pipe.add_operator(assay_op)

    def test_add_empty_pipeline_raises_error(self, pipe):
        sub_pipe = Pipeline(id="empty-pipe")
        with pytest.raises(PipelineConfigurationError):
            pipe.add_operator(sub_pipe)

    def test_add_pipeline_with_assay_operators_to_pipeline_with_sample_operator_raises_error(self, pipe):
        sample_op = helpers.DummyRoiExtractor(id="dummy-roi-extractor-id")
        pipe.add_operator(sample_op)

        assay_op = helpers.DummyDescriptorPatcher(id="dummy-descriptor-patcher-id")
        assay_pipe = Pipeline(id="assay-pipe")
        assay_pipe.add_operator(assay_op)
        with pytest.raises(PipelineConfigurationError):
            pipe.add_operator(assay_pipe)

    def test_validate_dataflow_on_sample_ok_if_first_operator_is_roi_extractor(self, pipe):
        pipe.add_operator(helpers.DummyRoiExtractor())
        pipe.validate_dataflow()

    def test_validate_dataflow_on_sample_pipeline_raises_error_if_first_operator_is_not_roi_extractor(self, pipe):
        with pytest.raises(PipelineConfigurationError):
            pipe.add_operator(helpers.DummyRoiTransformer())
            pipe.validate_dataflow()

    def test_validate_dataflow_ok_if_multiple_operators_are_in_valid_order(self, pipe):
        pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        pipe.add_operator(helpers.DummyRoiTransformer(id="op2"))
        pipe.add_operator(helpers.DummyFeatureExtractor(id="op3"))
        pipe.add_operator(helpers.DummyFeatureTransformer(id="op4"))
        pipe.validate_dataflow()

    def test_validate_dataflow_raise_error_if_multiple_operators_are_not_in_valid_order(self, pipe):
        pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        pipe.add_operator(helpers.DummyRoiTransformer(id="op2"))
        pipe.add_operator(helpers.DummyFeatureTransformer(id="op3"))
        pipe.add_operator(helpers.DummyFeatureExtractor(id="op4"))
        with pytest.raises(PipelineConfigurationError):
            pipe.validate_dataflow()

    def test_validate_dataflow_with_nested_pipeline_as_first_element(self, pipe):
        sub_pipe = Pipeline("sub-pipeline")
        sub_pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        sub_pipe.add_operator(helpers.DummyRoiTransformer(id="op2"))
        sub_pipe.add_operator(helpers.DummyFeatureExtractor(id="op3"))
        pipe.add_operator(sub_pipe)
        pipe.add_operator(helpers.DummyFeatureTransformer(id="op4"))
        pipe.validate_dataflow()

    def test_validate_dataflow_nested_pipeline(self, pipe):
        pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        sub_pipe = Pipeline("sub-pipeline")
        sub_pipe.add_operator(helpers.DummyRoiTransformer(id="op2"))
        sub_pipe.add_operator(helpers.DummyFeatureExtractor(id="op3"))
        sub_pipe.add_operator(helpers.DummyFeatureTransformer(id="op4"))
        pipe.add_operator(sub_pipe)
        pipe.validate_dataflow()

    def test_pipe_serialize_deserialize_flat_pipeline(self, pipe):
        pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        pipe.add_operator(helpers.DummyRoiTransformer(id="op2"))
        pipe.add_operator(helpers.DummyFeatureExtractor(id="op3"))
        pipe.add_operator(helpers.DummyFeatureTransformer(id="op4"))

        pipe_serialized_deserialized = Pipeline.deserialize(pipe.serialize())

        assert pipe == pipe_serialized_deserialized

    def test_pipe_serialize_deserialize_nested_pipeline(self, pipe):
        pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        sub_pipe = Pipeline(id="sub-pipe")
        sub_pipe.add_operator(helpers.DummyRoiTransformer(id="op2"))
        sub_pipe.add_operator(helpers.DummyFeatureExtractor(id="op3"))
        sub_pipe.add_operator(helpers.DummyFeatureTransformer(id="op4"))
        pipe.add_operator(sub_pipe)

        pipe_serialized_deserialized = Pipeline.deserialize(pipe.serialize())

        assert pipe == pipe_serialized_deserialized

    def test_deserialize_dict_without_id_raises_error(self):
        d = dict()
        with pytest.raises(ValueError):
            Pipeline.deserialize(d)

    def test_deserialize_dict_without_operators_raises_error(self):
        d = {"id": "pipeline-id"}
        with pytest.raises(ValueError):
            Pipeline.deserialize(d)

    def test_deserialize_dict_with_invalid_operator_list_raises_error(self):
        d = {"id": "pipeline-id", "operators": [1]}
        with pytest.raises(ValueError):
            Pipeline.deserialize(d)

    def test_apply_flat_pipeline(self, pipe, sample_storage):
        assert not sample_storage.get_n_rois()
        assert not sample_storage.get_n_features()
        pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        pipe.add_operator(helpers.DummyRoiTransformer(id="op2", max_length=10))
        pipe.add_operator(helpers.DummyFeatureExtractor(id="op3"))
        pipe.add_operator(helpers.DummyFeatureTransformer(id="op4"))
        pipe.apply(sample_storage)
        assert sample_storage.get_n_rois()
        assert sample_storage.get_n_features()

    def test_list_operators(self, pipe):
        pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        pipe.add_operator(helpers.DummyRoiTransformer(id="op2"))
        pipe.add_operator(helpers.DummyFeatureExtractor(id="op3"))
        pipe.add_operator(helpers.DummyFeatureTransformer(id="op4"))

        assert pipe.list_operator_ids() == ["op1", "op2", "op3", "op4"]

    def test_fetch_operator(self, pipe):
        query_op = helpers.DummyRoiTransformer(id="op2")
        pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        pipe.add_operator(query_op)
        pipe.add_operator(helpers.DummyFeatureExtractor(id="op3"))
        pipe.add_operator(helpers.DummyFeatureTransformer(id="op4"))

        fetched_op = pipe.get_operator(query_op.id)
        assert fetched_op == query_op

    def test_fetch_operator_not_found_raises_error(self, pipe):
        query_op = helpers.DummyRoiTransformer(id="op2")
        pipe.add_operator(helpers.DummyRoiExtractor(id="op1"))
        pipe.add_operator(query_op)
        pipe.add_operator(helpers.DummyFeatureExtractor(id="op3"))
        pipe.add_operator(helpers.DummyFeatureTransformer(id="op4"))

        with pytest.raises(ValueError):
            pipe.get_operator("invalid_id")



"""SQLite assay storage models."""

from __future__ import annotations

import json
from typing import Any
from uuid import UUID

from sqlalchemy import Float, Integer, String
from sqlalchemy.orm import Mapped, declarative_base, mapped_column

from ...core.dataflow import AssayProcessStatus
from ...core.models import Annotation, AnnotationPatch, DescriptorPatch, Feature, FeatureGroup, FillValue, Roi, Sample

Base = declarative_base()

SAMPLE_TABLE = "samples"
ROI_TABLE = "rois"
FEATURE_TABLE = "features"
ANNOTATION_TABLE = "annotations"
DESCRIPTOR_TABLE = "descriptors"


class BaseOrmModel(Base):
    """Base ORM model."""

    __abstract__ = True

    @classmethod
    def get_unique_columns(cls) -> list[str]:
        """Create a list of field names with unique constraint."""
        return [c.name for c in cls.__table__.c if c.unique]

    def to_dict(self) -> dict[str, Any]:
        """Convert model to a dictionary."""
        return {field.name: getattr(self, field.name) for field in self.__table__.c}

    def update(self, data: dict[str, Any]):
        """Update model using a dictionary."""
        for k, v in data.items():
            if k in self.__table__.c:
                setattr(self, k, v)


class BaseWithId(BaseOrmModel):
    """Base model with id field."""

    __abstract__ = True

    id: Mapped[UUID] = mapped_column(primary_key=True, doc="The record id")


class SampleModel(BaseOrmModel):
    """Stores data from :py:class:`tidyms2.core.models.Sample`."""

    __tablename__ = SAMPLE_TABLE

    id: Mapped[str] = mapped_column(String, primary_key=True)

    path: Mapped[str] = mapped_column(String, nullable=False)
    reader: Mapped[str] = mapped_column(String, nullable=True)
    ms_data_mode: Mapped[str] = mapped_column(String, nullable=True)
    ms_level: Mapped[int] = mapped_column(Integer)
    start_time: Mapped[float] = mapped_column(Float)
    end_time: Mapped[float | None] = mapped_column(Float, nullable=True)

    meta: Mapped[str] = mapped_column(String, nullable=True)

    @classmethod
    def from_sample(cls, sample: Sample) -> SampleModel:
        """Convert to pydantic model."""
        d = sample.model_dump(mode="json", exclude={"meta"})
        d["meta"] = sample.meta.model_dump_json()
        return SampleModel(**d)

    def to_pydantic_model(self) -> Sample:
        """Convert to pydantic model."""
        d = self.to_dict()
        if self.meta is not None:
            d["meta"] = json.loads(d["meta"])
        return Sample(**d)


class RoiModel(BaseWithId):
    """Stores data from :py:class:`tidyms2.core.models.Roi`."""

    __tablename__ = ROI_TABLE

    data: Mapped[str] = mapped_column(String, nullable=False)
    sample_id: Mapped[str] = mapped_column(String, nullable=False)
    snapshot_id: Mapped[str] = mapped_column(String, nullable=False)

    @classmethod
    def from_roi(cls, roi: Roi, snapshot_id: str) -> RoiModel:
        """Create a new instance from a pydantic model."""
        return cls(id=roi.id, data=roi.to_str(), sample_id=roi.sample.id, snapshot_id=snapshot_id)


class FeatureModel(BaseWithId):
    """Stores data from :py:class:`tidyms2.core.models.Roi`."""

    __tablename__ = FEATURE_TABLE

    data: Mapped[str] = mapped_column(String, nullable=False)
    roi_id: Mapped[UUID] = mapped_column(nullable=False)
    sample_id: Mapped[str] = mapped_column(String, nullable=False)
    snapshot_id: Mapped[str] = mapped_column(String, nullable=False)

    @classmethod
    def from_feature(cls, feature: Feature, snapshot_id: str) -> FeatureModel:
        """Create a new instance from a pydantic model."""
        return cls(
            id=feature.id,
            data=feature.to_str(),
            roi_id=feature.roi.id,
            sample_id=feature.roi.sample.id,
            snapshot_id=snapshot_id,
        )


class AnnotationModel(BaseWithId):
    """Stores data from :py:class:`tidyms2.core.models.Roi`."""

    __tablename__ = ANNOTATION_TABLE

    roi_id: Mapped[UUID] = mapped_column(nullable=False)
    sample_id: Mapped[str] = mapped_column(String, nullable=False)
    group: Mapped[int] = mapped_column(Integer, nullable=False)
    isotopologue_label: Mapped[int] = mapped_column(Integer, nullable=False)
    isotopologue_index: Mapped[int] = mapped_column(Integer, nullable=False)
    charge: Mapped[int] = mapped_column(Integer, nullable=False)

    snapshot_id: Mapped[str] = mapped_column(String, nullable=False)

    @classmethod
    def from_feature(cls, feature: Feature, snapshot_id: str) -> AnnotationModel:
        """Create a new instance from a pydantic model."""
        annotation: Annotation = feature.annotation  # type: ignore
        return cls(
            id=annotation.id,
            roi_id=annotation.roi_id,
            sample_id=annotation.sample_id,
            group=annotation.group,
            isotopologue_label=annotation.isotopologue_label,
            isotopologue_index=annotation.isotopologue_index,
            charge=annotation.charge,
            snapshot_id=snapshot_id,
        )

    def to_annotation(self) -> Annotation:
        """Convert model to annotation."""
        d = self.to_dict()
        d.pop("snapshot_id")
        return Annotation(**d)


class DescriptorModel(BaseWithId):
    """Stores data from :py:class:`tidyms2.core.models.Roi`."""

    __tablename__ = DESCRIPTOR_TABLE

    data: Mapped[str] = mapped_column(String, nullable=False)

    snapshot_id: Mapped[str] = mapped_column(String, nullable=False)

    @classmethod
    def from_feature(cls, feature: Feature, snapshot_id: str) -> DescriptorModel:
        """Create a new instance from a pydantic model."""
        return cls(id=feature.id, data=json.dumps(feature.describe()), snapshot_id=snapshot_id)

    def create_descriptor_dict(self) -> dict[str, float]:
        """Deserialize model into a dictionary of descriptors."""
        return json.loads(self.data)


class SampleSnapshotModel(BaseOrmModel):
    """Store sample snapshot ids and their order."""

    __tablename__ = "sample_snapshot_status"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    sample_id: Mapped[str] = mapped_column(String, nullable=False)
    order: Mapped[int] = mapped_column(Integer, nullable=False)
    snapshot_id: Mapped[str] = mapped_column(String, nullable=False)
    status: Mapped[str] = mapped_column(String, nullable=False)


class AssaySnapshotStatusModel(BaseOrmModel):
    """Store assay snapshot ids and their order."""

    __tablename__ = "assay_snapshot_status"

    order: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    id: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    status: Mapped[str] = mapped_column(String, nullable=False)

    def to_status(self) -> AssayProcessStatus:
        """Convert model into an assay process state instance."""
        return AssayProcessStatus.model_validate_json(self.status)


class AnnotationPatchModel(BaseOrmModel):
    """Store patches applied to annotation data."""

    __tablename__ = "annotation_patches"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    feature_id: Mapped[UUID] = mapped_column(nullable=False)
    field: Mapped[str] = mapped_column(String, nullable=False)
    value: Mapped[int] = mapped_column(Integer, nullable=False)
    snapshot_id: Mapped[str] = mapped_column(String, nullable=False)

    @classmethod
    def from_patch(cls, patch: AnnotationPatch, snapshot: str) -> AnnotationPatchModel:
        """Create a new instance from an annotation patch."""
        return AnnotationPatchModel(feature_id=patch.id, field=patch.field, value=patch.value, snapshot_id=snapshot)


class DescriptorPatchModel(BaseOrmModel):
    """Store patches applied to descriptor data."""

    __tablename__ = "descriptor_patches"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    feature_id: Mapped[UUID] = mapped_column(nullable=False)
    descriptor: Mapped[str] = mapped_column(String, nullable=False)
    value: Mapped[float] = mapped_column(Float, nullable=False)
    snapshot_id: Mapped[str] = mapped_column(String, nullable=False)

    @classmethod
    def from_patch(cls, patch: DescriptorPatch, snapshot: str) -> AnnotationPatchModel:
        """Create a new instance from an annotation patch."""
        return DescriptorPatchModel(
            feature_id=patch.id, descriptor=patch.descriptor, value=patch.value, snapshot_id=snapshot
        )


class FillValueModel(BaseOrmModel):
    """Store patches applied to annotation data."""

    __tablename__ = "fill_values"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    feature_group: Mapped[int] = mapped_column(nullable=False)
    sample_id: Mapped[str] = mapped_column(String, nullable=False)
    value: Mapped[float] = mapped_column(Float, nullable=False)
    snapshot_id: Mapped[str] = mapped_column(String, nullable=False)

    @classmethod
    def from_fill_value(cls, fill: FillValue, snapshot: str) -> FillValueModel:
        """Create a new instance from an annotation patch."""
        return FillValueModel(
            feature_group=fill.feature_group, sample_id=fill.sample_id, value=fill.value, snapshot_id=snapshot
        )

    def to_fill_value(self) -> FillValue:
        """Create a fill value pydantic model from the current instance."""
        return FillValue(sample_id=self.sample_id, feature_group=self.feature_group, value=self.value)


class FeatureGroupModel(BaseOrmModel):
    """Store feature group data."""

    __tablename__ = "feature_groups"

    group: Mapped[int] = mapped_column(Integer, primary_key=True)
    descriptors: Mapped[str] = mapped_column(nullable=False)
    annotations: Mapped[str] = mapped_column(nullable=False)

    @classmethod
    def from_feature_group(cls, feature_group: FeatureGroup):
        """Create a new model from a feature group."""
        return cls(
            group=feature_group.group,
            descriptors=json.dumps(feature_group.descriptors),
            annotations=feature_group.annotation.model_dump_json(),
        )

    def to_feature_group(self) -> FeatureGroup:
        """Convert a model into a feature group."""
        return FeatureGroup.model_validate(
            {
                "group": self.group,
                "annotation": json.loads(self.annotations),
                "descriptors": json.loads(self.descriptors),
            }
        )



"""LC-MS assay operators."""

from typing import TYPE_CHECKING

import pydantic

from ...annotation.correspondence import FeatureCorrespondenceParameters, match_features
from ...core.enums import MSInstrument, Polarity, SeparationMode
from ...core.models import AnnotationPatch, MZTrace
from ...core.operators.assay import AnnotationPatcher
from ...core.storage import AssayStorage
from ..models import Peak

if TYPE_CHECKING:
    from typing import assert_never


class LCFeatureMatcher(AnnotationPatcher, FeatureCorrespondenceParameters):
    """Perform feature correspondence on LC-MS datasets using a cluster-based approach.

    Features are initially grouped by m/z and Rt similarity using DBSCAN. In a second step, these
    clusters are further processed using a GMM approach, obtaining clusters where each sample
    contributes with only one sample.

    See the :ref:`algorithms-correspondence` guide for a detailed description of the algorithm.

    """

    mz_tolerance: pydantic.PositiveFloat = 0.01
    """m/z tolerance used to group close features. Sets the `eps` parameter in the DBSCAN algorithm."""

    rt_tolerance: pydantic.PositiveFloat = 3.0
    """Rt tolerance in seconds used to group close features. Sets the `eps` parameter in the DBSCAN algorithm."""

    def compute_patches(self, data: AssayStorage[MZTrace, Peak]) -> list[AnnotationPatch]:
        """Compute annotation patches for feature matching."""
        tolerance = {"mz": self.mz_tolerance, "rt": self.rt_tolerance}
        descriptors = data.fetch_descriptors(descriptors=list(tolerance))
        annotations = data.fetch_annotations()
        samples = data.list_samples()
        return match_features(descriptors, annotations, samples, tolerance, self)

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity):
        """Set the processor default parameters.

        :param instrument : the instrument type used in the experimental setup
        :param separation : the LC platform used in the experimental setup
        :param polarity : the MS polarity used in the experiment

        """
        op = cls()

        match instrument:
            case MSInstrument.QTOF:
                op.mz_tolerance = 0.01
            case MSInstrument.ORBITRAP:
                op.mz_tolerance = 0.005
            case _ as never:
                assert_never(never)

        match separation:
            case SeparationMode.HPLC:
                op.rt_tolerance = 10.0
            case SeparationMode.UPLC:
                op.rt_tolerance = 5.0
            case SeparationMode.DART:
                raise NotImplementedError
            case _ as never:
                assert_never(never)
        return op



"""Abstract matrix operators."""

from __future__ import annotations

import concurrent.futures
from abc import abstractmethod
from multiprocessing import get_context

from ..dataflow import DataMatrixProcessStatus
from ..matrix import DataMatrix, FeatureVector, SampleVector
from ..utils.numpy import FloatArray
from .base import BaseOperator


class MatrixOperator(BaseOperator[DataMatrixProcessStatus]):
    """Base operator for data matrix transformation."""

    def apply(self, data: DataMatrix) -> None:
        """Apply the operator function to the data."""
        self.check_status(data.get_process_status())

        if hasattr(self, "pre_apply"):
            self.pre_apply(data)  # type: ignore

        self._apply_operator(data)

        if hasattr(self, "post_apply"):
            self.post_apply(data)  # type: ignore

        self.update_status(data.get_process_status())

    @abstractmethod
    def _apply_operator(self, data: DataMatrix) -> None: ...


class MatrixTransformer(MatrixOperator):
    """Apply an arbitrary transformation to a data matrix.

    MUST implement `transform_matrix` which takes a data matrix and produces a 2D array
    that will replace the matrix data.

    """

    def _apply_operator(self, data: DataMatrix) -> None:
        data.set_data(self._transform_matrix(data))

    @abstractmethod
    def _transform_matrix(self, data: DataMatrix) -> FloatArray: ...


class ColumnFilter(MatrixOperator):
    """Remove data matrix features based on a specific condition.

    MUST implement the method `_create_remove_list` that takes the data matrix and produce a
    list of feature groups to remove.

    """

    exclude: list[int] | None = None
    """A list of feature groups to ignore during filtering."""

    def _apply_operator(self, data: DataMatrix) -> None:
        exclude = self.exclude or set()
        data.remove_features(*(g for g in self._create_remove_list(data) if g not in exclude))

    @abstractmethod
    def _create_remove_list(self, data: DataMatrix) -> list[int]: ...


class RowFilter(MatrixOperator):
    """Remove data matrix samples based on a specific condition.

    MUST implement the method `_create_remove_list` that takes the data matrix and produce a
    list of feature groups to remove.

    """

    exclude: list[str] | None = None
    """A list of sample ids to ignore during filtering."""

    def _apply_operator(self, data: DataMatrix) -> None:
        exclude = set() if self.exclude is None else set(self.exclude)
        data.remove_samples(*(g for g in self._create_remove_list(data) if g not in exclude))

    @abstractmethod
    def _create_remove_list(self, data: DataMatrix) -> list[str]: ...


class MultiProcessTransformer(MatrixOperator):
    """Abstract operator with configuration for using multiprocess transformations."""

    max_workers: int | None = None
    """the maximum number of workers subprocesses"""

    spawn_context: str = "spawn"
    """the method used to create subprocesses, passed to :py:func:`multiprocess.get_context`"""


class ColumnTransformer(MultiProcessTransformer):
    """Transform a data matrix columns based on an arbitrary transformation.

    MUST implement the method `_transform_column` that takes a feature vector and creates a
    transformed vector.

    """

    exclude: list[int] | None = None
    """A list of feature groups to ignore during filtering."""

    def _apply_operator(self, data: DataMatrix) -> None:
        exclude = set() if self.exclude is None else set(self.exclude)
        include = [x.group for x in data.features if x.group not in exclude]

        transformed: list[FeatureVector] = list()
        mp_context = get_context(method=self.spawn_context)
        with concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers, mp_context=mp_context) as executor:
            futures = [executor.submit(self._transform_column, s) for s in data.get_columns(*include)]
            for fs in concurrent.futures.as_completed(futures):
                transformed.append(fs.result())

        data.set_columns(*((x.feature.group, x.data) for x in transformed))

    @abstractmethod
    def _transform_column(self, column: FeatureVector) -> FeatureVector: ...


class RowTransformer(MultiProcessTransformer):
    """Transform a data matrix rows based on an arbitrary transformation.

    MUST implement the method `_transform_row` that takes a sample vector and creates a
    transformed vector.

    """

    exclude: list[str] | None = None
    """A list of sample ids to exclude from the transformation"""

    def _apply_operator(self, data: DataMatrix) -> None:
        exclude = set() if self.exclude is None else set(self.exclude)
        include = [x.id for x in data.samples if x.id not in exclude]

        transformed: list[SampleVector] = list()
        mp_context = get_context(method=self.spawn_context)
        with concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers, mp_context=mp_context) as executor:
            futures = [executor.submit(self._transform_row, s) for s in data.get_rows(*include)]
            for fs in concurrent.futures.as_completed(futures):
                transformed.append(fs.result())

        data.set_rows(*((x.sample.id, x.data) for x in transformed))

    @abstractmethod
    def _transform_row(self, row: SampleVector) -> SampleVector: ...



import pytest

from tidyms2.annotation import consensus
from tidyms2.annotation.consensus import AnnotationGraph, MissingNodeError
from tidyms2.core.models import Annotation
from tidyms2.core.utils.common import create_id


def create_annotation(**kwargs) -> Annotation:
    """Create an annotation with dummy feature and ROI ids.

    :param kwargs: key/value arguments passed to the Annotation constructor
    """
    return Annotation(id=create_id(), roi_id=create_id(), **kwargs)


class TestAnnotationGraphNodeFunctionality:
    @pytest.fixture
    def annotation(self):
        return create_annotation(group=1, charge=1, isotopologue_index=1)

    def test_add_single_annotation(self, annotation):
        graph = AnnotationGraph()
        graph.add_annotation(annotation)

        node = graph.get_node(annotation.group)

        assert node.group == annotation.group
        assert len(graph.list_node_groups()) == 1
        assert node.charge_votes[annotation.group] == 1
        assert node.index_votes[annotation.group] == 1

    def test_add_multiple_times_annotation_with_same_group_updates_same_node(self, annotation: Annotation):
        graph = AnnotationGraph()
        n_times_added = 5
        for _ in range(n_times_added):
            graph.add_annotation(annotation)

        node = graph.get_node(annotation.group)

        assert node.group == annotation.group
        assert len(graph.list_node_groups()) == 1
        assert node.charge_votes[annotation.charge] == n_times_added
        assert node.index_votes[annotation.isotopologue_index] == n_times_added

    def test_add_multiple_annotation_with_equal_group_and_different_charge_state_update_votes(
        self, annotation: Annotation
    ):
        annotation2 = annotation.model_copy()
        annotation2.charge = 2

        graph = AnnotationGraph()
        graph.add_annotation(annotation)
        graph.add_annotation(annotation2)

        assert len(graph.list_node_groups()) == 1

        node = graph.get_node(annotation.group)
        assert node.group == annotation.group
        # both annotations have different charge but equal isotopologue index
        assert node.charge_votes[annotation.charge] == 1
        assert node.charge_votes[annotation2.charge] == 1
        assert node.index_votes[annotation.isotopologue_index] == 2

    def test_add_annotation_with_different_group_creates_new_nodes(self, annotation: Annotation):
        annotation2 = annotation.model_copy()
        annotation2.group = 2
        annotation2.charge = 2

        graph = AnnotationGraph()
        graph.add_annotation(annotation)
        graph.add_annotation(annotation2)

        assert len(graph.list_node_groups()) == 2

        node1 = graph.get_node(annotation.group)
        assert node1.group == annotation.group
        assert node1.charge_votes[annotation.charge] == 1
        assert node1.index_votes[annotation.isotopologue_index] == 1

        node2 = graph.get_node(annotation2.group)
        assert node2.group == annotation2.group
        assert node2.charge_votes[annotation2.charge] == 1
        assert node2.index_votes[annotation2.isotopologue_index] == 1

    def test_get_node_missing_node_raises_error(self):
        graph = AnnotationGraph()
        with pytest.raises(MissingNodeError):
            graph.get_node(1000)

    def test_remove_node_ok(self, annotation: Annotation):
        graph = AnnotationGraph()

        # add node and check that it is retrievable
        graph.add_annotation(annotation)
        graph.get_node(annotation.group)

        graph.remove_node(annotation.group)
        with pytest.raises(MissingNodeError):
            graph.get_node(annotation.group)

    def test_count_votes(self, annotation: Annotation):
        # add multiple times an annotation with same group but different
        # charge state. `charge` field should be set to 2 after counting votes

        graph = AnnotationGraph()

        votes_with_charge_one = 2
        for _ in range(votes_with_charge_one):
            graph.add_annotation(annotation)

        expected_node_charge = 2
        annotation.charge = expected_node_charge
        votes_with_charge_two = 3

        for _ in range(votes_with_charge_two):
            graph.add_annotation(annotation)

        graph.count_votes()

        node = graph.get_node(annotation.group)
        assert node.charge == expected_node_charge


class TestAnnotationGraphEdgeFunctionality:
    @pytest.fixture
    def graph(self):
        g = AnnotationGraph()

        for group in range(3):
            ann = create_annotation(group=group, charge=1, isotopologue_index=0)
            g.add_annotation(ann)
        return g

    def test_add_single_edge(self, graph: AnnotationGraph):
        group1, group2 = 0, 1
        graph.add_edge(group1, group2)

        node1_edges = graph.get_edges(group1)
        assert len(node1_edges) == 1
        edge1 = node1_edges[0]
        assert edge1.groups == (group1, group2)
        assert edge1.votes == 1

        node2_edges = graph.get_edges(group2)
        assert len(node2_edges) == 1
        edge2 = node2_edges[0]
        assert edge2.groups == (group2, group1)
        assert edge2.votes == 1

    def test_add_multiple_times_same_edge(self, graph: AnnotationGraph):
        group1, group2 = 0, 1
        add_count = 5
        for _ in range(add_count):
            graph.add_edge(group1, group2)

        node1_edges = graph.get_edges(group1)
        assert len(node1_edges) == 1
        edge = node1_edges[0]
        assert edge.groups == (group1, group2)
        assert edge.votes == add_count

        node2_edges = graph.get_edges(group2)
        assert len(node2_edges) == 1
        edge2 = node2_edges[0]
        assert edge2.groups == (group2, group1)
        assert edge2.votes == add_count

    def test_remove_edges_ok(self, graph: AnnotationGraph):
        group1, group2 = 0, 1
        graph.add_edge(group1, group2)

        assert len(graph.get_edges(group1)) == 1
        assert len(graph.get_edges(group2)) == 1

        graph.remove_edge(group1, group2)

        assert len(graph.get_edges(group1)) == 0
        assert len(graph.get_edges(group2)) == 0

    def test_remove_edges_non_existing_edge_raises_error(self, graph: AnnotationGraph):
        with pytest.raises(ValueError):
            graph.remove_edge(0, 1)

    def test_no_neighbors_returns_empty_list(self, graph: AnnotationGraph):
        neighbors = graph.get_neighbors(0)
        assert len(neighbors) == 0

    def test_get_neighbors_ok(self, graph: AnnotationGraph):
        neighbor_groups = [1, 2]
        node_group = 0
        for neighbor_group in neighbor_groups:
            graph.add_edge(node_group, neighbor_group)

        neighbors = graph.get_neighbors(node_group)

        assert len(neighbors) == len(neighbor_groups)
        assert neighbors[0] == graph.get_node(neighbor_groups[0])
        assert neighbors[1] == graph.get_node(neighbor_groups[1])

    def test_remove_isolated_in_graph_with_no_edges_delete_all_nodes(self, graph: AnnotationGraph):
        graph.remove_isolated()
        all_node_groups = graph.list_node_groups()
        assert not all_node_groups

    def test_remove_isolated_ok(self, graph: AnnotationGraph):
        graph.add_edge(0, 1)
        graph.remove_isolated()

        all_node_groups = graph.list_node_groups()
        assert 0 in all_node_groups
        assert 1 in all_node_groups
        assert 2 not in all_node_groups


class TestAddSampleAnnotations:
    def test_annotation_with_no_group_are_not_added_to_the_graph(self):
        graph = AnnotationGraph()
        annotations = [create_annotation(isotopologue_index=0), create_annotation(isotopologue_index=0, group=1)]
        consensus.add_sample_annotations(graph, annotations)

        assert -1 not in graph.list_node_groups()

    def test_annotation_with_no_isotopologue_index_are_not_added_to_the_graph(self):
        graph = AnnotationGraph()
        exclude_group = 0
        annotations = [create_annotation(group=exclude_group), create_annotation(isotopologue_index=0, group=1)]
        consensus.add_sample_annotations(graph, annotations)

        assert exclude_group not in graph.list_node_groups()

    def test_non_zero_isotopologue_indices_only_neighbor_is_their_mmi(self):
        graph = AnnotationGraph()
        annotations = [
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1),
            create_annotation(group=2, isotopologue_label=0, isotopologue_index=2),
            create_annotation(group=3, isotopologue_label=1, isotopologue_index=0),
            create_annotation(group=4, isotopologue_label=2, isotopologue_index=1),
        ]
        consensus.add_sample_annotations(graph, annotations)

        # check isotopologue label == 0 mmi
        group0_neighbors = [x.group for x in graph.get_neighbors(0)]
        assert len(group0_neighbors) == 2
        assert 1 in group0_neighbors
        assert 2 in group0_neighbors

        # check isotopologue label == 0 non mmi
        group1_neighbors = [x.group for x in graph.get_neighbors(1)]
        assert group1_neighbors == [0]

        group2_neighbors = [x.group for x in graph.get_neighbors(2)]
        assert group2_neighbors == [0]

        # check isotopologue label == 0 mmi
        group3_neighbors = [x.group for x in graph.get_neighbors(3)]
        assert not group3_neighbors


class TestRemoveIncompatibleNeighbors:
    def test_isotopologue_neighbor_with_equal_index_is_removed(self):
        graph = AnnotationGraph()
        annotations = [
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample1"),
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample2"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample1"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample2"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample3"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample4"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample5"),
        ]
        for sample_annotations in consensus.group_annotations_by_sample(annotations).values():
            consensus.add_sample_annotations(graph, sample_annotations)
        graph.count_votes()

        mmi = graph.get_node(0)

        assert 1 in mmi.neighbors

        consensus.remove_invalid_mmi_neighbors(graph, mmi)

        assert 1 not in mmi.neighbors

    def test_isotopologue_neighbor_with_different_charge_is_removed(self):
        graph = AnnotationGraph()
        annotations = [
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample1"),
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample2"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample1"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample2"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=2, sample_id="sample3"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=2, sample_id="sample4"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=2, sample_id="sample5"),
        ]
        for sample_annotations in consensus.group_annotations_by_sample(annotations).values():
            consensus.add_sample_annotations(graph, sample_annotations)
        graph.count_votes()

        mmi = graph.get_node(0)

        assert 1 in mmi.neighbors

        consensus.remove_invalid_mmi_neighbors(graph, mmi)

        assert 1 not in mmi.neighbors

    def test_valid_neighbor_is_kept(self):
        graph = AnnotationGraph()
        annotations = [
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample1"),
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample2"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample1"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample2"),
        ]
        for sample_annotations in consensus.group_annotations_by_sample(annotations).values():
            consensus.add_sample_annotations(graph, sample_annotations)

        graph.count_votes()

        mmi = graph.get_node(0)

        assert 1 in mmi.neighbors

        consensus.remove_invalid_mmi_neighbors(graph, mmi)

        assert 1 in mmi.neighbors


class TestSolveMMIConflicts:
    def test_remove_isotopologue_with_least_votes(self):
        graph = AnnotationGraph()
        annotations = [
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample1"),
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample2"),
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample3"),
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample4"),
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample5"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample1"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample2"),
            create_annotation(group=2, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample3"),
            create_annotation(group=2, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample4"),
            create_annotation(group=2, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample5"),
        ]
        for sample_annotations in consensus.group_annotations_by_sample(annotations).values():
            consensus.add_sample_annotations(graph, sample_annotations)
        graph.count_votes()

        mmi = graph.get_node(0)

        assert 1 in mmi.neighbors
        assert 2 in mmi.neighbors

        consensus.solve_mmi_conflict(graph, mmi)

        assert 1 not in mmi.neighbors
        assert 2 in mmi.neighbors

    def test_all_valid_isotopologue_are_kept(self):
        graph = AnnotationGraph()
        annotations = [
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample1"),
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample2"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample1"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample2"),
            create_annotation(group=2, isotopologue_label=0, isotopologue_index=2, charge=1, sample_id="sample1"),
            create_annotation(group=2, isotopologue_label=0, isotopologue_index=2, charge=1, sample_id="sample2"),
        ]
        for sample_annotations in consensus.group_annotations_by_sample(annotations).values():
            consensus.add_sample_annotations(graph, sample_annotations)
        graph.count_votes()

        mmi = graph.get_node(0)

        assert 1 in mmi.neighbors
        assert 2 in mmi.neighbors

        consensus.solve_mmi_conflict(graph, mmi)

        assert 1 in mmi.neighbors
        assert 2 in mmi.neighbors


class TestSolveNonMMIConflict:
    def test_only_most_voted_mmi_neighbor_is_kept(self):
        graph = AnnotationGraph()
        annotations = [
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample1"),
            create_annotation(group=0, isotopologue_label=0, isotopologue_index=0, charge=1, sample_id="sample2"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample1"),
            create_annotation(group=1, isotopologue_label=0, isotopologue_index=1, charge=1, sample_id="sample2"),
            create_annotation(group=1, isotopologue_label=1, isotopologue_index=1, charge=1, sample_id="sample3"),
            create_annotation(group=2, isotopologue_label=1, isotopologue_index=0, charge=1, sample_id="sample3"),
        ]
        for sample_annotations in consensus.group_annotations_by_sample(annotations).values():
            consensus.add_sample_annotations(graph, sample_annotations)
        graph.count_votes()

        m1 = graph.get_node(1)

        assert 0 in m1.neighbors
        assert 2 in m1.neighbors

        consensus.solve_mmi_conflict(graph, m1)

        assert 0 in m1.neighbors
        assert 2 not in m1.neighbors



"""Chemistry related exceptions."""


class IsotopeError(ValueError):
    """Exception raised when an invalid isotope string is passed."""


class InvalidFormula(ValueError):
    """Exception raised when an invalid formula string is passed."""



import pytest

from tidyms2.core import exceptions
from tidyms2.core.dataflow import SampleProcessStatus
from tidyms2.core.models import AnnotationPatch, DescriptorPatch, FillValue
from tidyms2.core.utils.common import create_id
from tidyms2.storage import memory

from .. import helpers


class TestOnMemorySampleStorage:
    @pytest.fixture(scope="class")
    def data_path(self, tmp_path_factory):
        return tmp_path_factory.mktemp("test-on-memory-sample-storage-snapshot")

    @pytest.fixture(scope="class")
    def sample(self, data_path):
        return helpers.create_sample(data_path, 1)

    @pytest.fixture(scope="class")
    def roi1(self, sample):
        return helpers.create_roi(sample)

    @pytest.fixture(scope="class")
    def roi2(self, sample):
        return helpers.create_roi(sample)

    @pytest.fixture(scope="class")
    def ft11(self, roi1):
        return helpers.create_feature(roi1)

    @pytest.fixture(scope="class")
    def ft21(self, roi2):
        return helpers.create_feature(roi2)

    @pytest.fixture(scope="class")
    def ft22(self, roi2):
        return helpers.create_feature(roi2)

    def test_create(self, sample):
        memory.OnMemorySampleStorage(sample, helpers.ConcreteRoi, helpers.ConcreteFeature)

    @pytest.fixture
    def storage(self, sample):
        return memory.OnMemorySampleStorage(sample, helpers.ConcreteRoi, helpers.ConcreteFeature)

    def test_get_roi_invalid_id_raises_error(self, storage):
        with pytest.raises(exceptions.RoiNotFound):
            invalid_id = create_id()
            storage.get_roi(invalid_id)

    def test_has_roi_invalid_roi_id_raises_error(self, storage, roi1):
        invalid_id = create_id()
        assert not storage.has_roi(invalid_id)

    def test_has_roi_ok(self, storage, roi1):
        storage.add_rois(roi1)
        assert storage.has_roi(roi1.id)

    def test_get_roi(self, storage, roi1):
        expected = roi1
        storage.add_rois(expected)
        actual = storage.get_roi(expected.id)
        assert actual == expected

    def test_add_multiple_roi(self, storage, roi1, roi2):
        storage.add_rois(roi1, roi2)
        assert storage.has_roi(roi1.id)
        assert storage.has_roi(roi2.id)

    def test_add_rois_with_repeated_ids_does_not_add_any_roi(self, storage, roi1, roi2):
        storage.add_rois(roi1)
        with pytest.raises(exceptions.RepeatedIdError):
            storage.add_rois(roi2, roi1)
            assert storage.has_roi(roi1.id)
            assert not storage.has_roi(roi2.id)

    def test_add_roi_with_features_also_does_not_add_features(self, storage, roi1, ft11):
        storage.add_rois(roi1)
        assert storage.has_roi(roi1.id)
        assert not storage.has_feature(ft11.id)

    def test_list_rois_no_roi_returns_empty_list(self, storage):
        result = storage.list_rois()
        assert not result

    def test_list_rois_ok(self, storage, roi1, roi2):
        storage.add_rois(roi1, roi2)
        result = storage.list_rois()
        assert roi1 in result
        assert roi2 in result
        assert len(result) == 2

    def test_delete_single_roi(self, storage, roi1):
        storage.add_rois(roi1)
        assert storage.has_roi(roi1.id)

        storage.delete_rois(roi1.id)
        assert not storage.has_roi(roi1.id)

    def test_delete_multiple_rois(self, storage, roi1, roi2):
        storage.add_rois(roi1)
        storage.add_rois(roi2)

        storage.delete_rois(roi1.id, roi2.id)
        assert not storage.has_roi(roi1.id)
        assert not storage.has_roi(roi2.id)

    def test_delete_non_existing_roi_id_is_ignored(self, storage, roi1, roi2):
        storage.add_rois(roi1)

        storage.delete_rois(roi2.id)

        assert storage.has_roi(roi1.id)
        assert not storage.has_roi(roi2.id)

    def test_delete_roi_with_features_also_delete_features(self, storage, roi1, ft11):
        storage.add_rois(roi1)
        storage.add_features(ft11)

        assert storage.has_roi(roi1.id)
        assert storage.has_feature(ft11.id)

        storage.delete_rois(roi1.id)

        assert not storage.has_roi(roi1.id)
        assert not storage.has_feature(ft11.id)

    def test_add_single_feature(self, storage, roi1):
        storage.add_rois(roi1)

        # check that id is not on storage before creating
        feature_id = create_id()
        assert not storage.has_feature(feature_id)

        # check that feature is not added on creation
        feature = helpers.ConcreteFeature(id=feature_id, roi=roi1, data_mz=1)
        assert not storage.has_feature(feature_id)

        # check ok after add
        storage.add_features(feature)
        assert storage.has_feature(feature_id)

    def test_add_feature_raise_error_if_parent_roi_is_not_stored(self, storage, ft11):
        with pytest.raises(exceptions.RoiNotFound):
            storage.add_features(ft11)

    def test_add_multiple_features(self, storage, roi1):
        storage.add_rois(roi1)

        ft1 = helpers.ConcreteFeature(id=create_id(), roi=roi1, data_mz=1)
        ft2 = helpers.ConcreteFeature(id=create_id(), roi=roi1, data_mz=1)

        assert not storage.has_feature(ft1.id)
        assert not storage.has_feature(ft2.id)

        storage.add_features(ft1, ft2)

        assert storage.has_feature(ft1.id)
        assert storage.has_feature(ft2.id)

    def test_add_feature_with_repeated_id_raises_error(self, storage, roi1, ft11):
        storage.add_rois(roi1)
        storage.add_features(ft11)
        with pytest.raises(exceptions.RepeatedIdError):
            storage.add_features(ft11)

    def test_add_feature_with_parent_roi_not_stored_raises_error(self, storage, roi1, ft11):
        with pytest.raises(exceptions.RoiNotFound):
            storage.add_features(ft11)

    def test_get_feature(self, storage, roi1, ft11):
        storage.add_rois(roi1)
        storage.add_features(ft11)

        expected = ft11
        actual = storage.get_feature(expected.id)

        assert actual == expected

    def test_get_feature_non_existing_id_raises_error(self, storage):
        with pytest.raises(exceptions.FeatureNotFound):
            id_ = create_id()
            storage.get_feature(id_)

    def test_features_no_feature_return_empty_list(self, storage):
        assert not storage.list_features()

    def test_features_all_rois(self, storage, roi1, roi2, ft11, ft21):
        storage.add_rois(roi1, roi2)
        storage.add_features(ft11, ft21)
        features = storage.list_features()
        assert ft11 in features
        assert ft21 in features

    def test_features_single_roi(self, storage, roi1, roi2, ft11, ft21):
        storage.add_rois(roi1, roi2)
        storage.add_features(ft11, ft21)
        features = storage.list_features(roi_id=roi1.id)
        assert ft11 in features
        assert ft21 not in features

    def test_features_non_existing_roi_id_raises_error(self, storage, roi1):
        with pytest.raises(exceptions.RoiNotFound):
            storage.list_features(roi_id=roi1.id)

    def test_delete_single_feature(self, storage, roi1, ft11):
        storage.add_rois(roi1)
        storage.add_features(ft11)

        storage.delete_features(ft11.id)

        assert not storage.has_feature(ft11.id)

    def test_delete_multiple_features(self, storage, roi1, roi2, ft11, ft21, ft22):
        storage.add_rois(roi1, roi2)
        storage.add_features(ft11, ft21, ft22)

        storage.delete_features(ft11.id, ft21.id)

        assert not storage.has_feature(ft11.id)
        assert not storage.has_feature(ft21.id)
        assert storage.has_feature(ft22.id)

    def test_delete_features_non_existing_ids_are_ignored(self, storage, roi1, roi2, ft11, ft21):
        storage.add_rois(roi1, roi2)
        storage.add_features(ft11, ft21)

        non_existing_id = create_id()

        storage.delete_features(non_existing_id)

        assert storage.has_feature(ft11.id)
        assert storage.has_feature(ft21.id)

    @pytest.fixture
    def storage_with_snapshot(self, sample, roi1, ft11):
        storage = memory.OnMemorySampleStorage(sample, helpers.ConcreteRoi, helpers.ConcreteFeature)
        storage.add_rois(roi1)
        storage.add_features(ft11)
        storage.create_snapshot("snapshot")
        return storage

    def test_create_snapshot_with_reserved_id_raises_error(self, storage):
        with pytest.raises(ValueError):
            storage.create_snapshot(memory.LATEST)

    def test_create_snapshot_sets_current_snapshot_as_latest(self, storage_with_snapshot):
        storage_with_snapshot.create_snapshot("new-snapshot")
        assert storage_with_snapshot.get_snapshot_id() == memory.LATEST

    def test_list_snapshots_no_snapshots_on_creation(self, storage):
        assert storage.list_snapshots() == []

    def test_list_snapshots_on_storage_with_data_but_no_snapshots_has_only_latest_snapshot_id(
        self, storage, roi1, roi2
    ):
        storage.add_rois(roi1, roi2)
        assert storage.list_snapshots() == [memory.LATEST]

    def test_list_snapshots(self, storage_with_snapshot):
        expected = ["snapshot", memory.LATEST]
        actual = storage_with_snapshot.list_snapshots()
        assert actual == expected

    def test_create_snapshot_with_existing_id_raises_error(self, storage_with_snapshot):
        with pytest.raises(exceptions.RepeatedIdError):
            storage_with_snapshot.create_snapshot("snapshot")

    def test_set_snapshot_non_existing_id_raises_error(self, storage_with_snapshot):
        with pytest.raises(exceptions.SnapshotNotFound):
            storage_with_snapshot.set_snapshot("invalid-snapshot-id")

    def test_add_roi_and_set_to_snapshot_rollback_changes(self, storage_with_snapshot, roi2):
        storage_with_snapshot.add_rois(roi2)

        assert storage_with_snapshot.has_roi(roi2.id)

        storage_with_snapshot.set_snapshot("snapshot")

        assert not storage_with_snapshot.has_roi(roi2.id)

    def test_add_feature_and_set_to_snapshot_rollback_changes(self, storage_with_snapshot, roi2, ft21):
        storage_with_snapshot.add_rois(roi2)
        storage_with_snapshot.add_features(ft21)

        assert storage_with_snapshot.has_feature(ft21.id)

        storage_with_snapshot.set_snapshot("snapshot")

        assert not storage_with_snapshot.has_feature(ft21.id)

    def test_try_to_add_roi_to_non_latest_snapshot_raises_error(self, storage_with_snapshot):
        storage_with_snapshot.set_snapshot("snapshot")
        sample = storage_with_snapshot.get_sample()
        dummy_roi = helpers.create_roi(sample)
        with pytest.raises(exceptions.SnapshotError):
            storage_with_snapshot.add_rois(dummy_roi)

    def test_try_to_add_feature_to_non_latest_snapshot_raises_error(self, storage_with_snapshot):
        storage_with_snapshot.set_snapshot("snapshot")
        sample = storage_with_snapshot.get_sample()
        dummy_roi = helpers.create_roi(sample)
        dummy_ft = helpers.create_feature(dummy_roi)
        with pytest.raises(exceptions.SnapshotError):
            storage_with_snapshot.add_features(dummy_ft)

    def test_try_to_set_non_latest_snapshot_status_raises_error(self, storage_with_snapshot):
        storage_with_snapshot.set_snapshot("snapshot")
        new_status = SampleProcessStatus(feature_extracted=True)
        with pytest.raises(exceptions.SnapshotError):
            storage_with_snapshot.set_status(new_status)

    def test_set_status(self, storage):
        expected = SampleProcessStatus(isotopologue_annotated=True)
        storage.set_status(expected)
        actual = storage.get_status()
        assert actual == expected

    def test_set_snapshot_with_reset(self, storage_with_snapshot):
        storage_with_snapshot.set_snapshot("snapshot", reset=True)

        assert "snapshot" not in storage_with_snapshot.list_snapshots()

        with pytest.raises(exceptions.SnapshotNotFound):
            storage_with_snapshot.set_snapshot("snapshot")

    def test_using_snapshot_changes_roi_id(self, storage_with_snapshot):
        storage_with_snapshot.set_snapshot("snapshot")
        roi_snapshot = storage_with_snapshot.list_rois()[0]
        storage_with_snapshot.set_snapshot()
        roi_latest = storage_with_snapshot.list_rois()[0]

        assert roi_snapshot.id != roi_latest.id
        assert roi_latest.equal(roi_snapshot)

    def test_using_snapshot_changes_feature_id(self, storage_with_snapshot):
        storage_with_snapshot.set_snapshot("snapshot")
        feature_snapshot = storage_with_snapshot.list_features()[0]
        storage_with_snapshot.set_snapshot()
        feature_latest = storage_with_snapshot.list_features()[0]

        assert feature_snapshot.id != feature_latest.id
        assert feature_snapshot.roi.id != feature_latest.roi.id
        assert feature_latest.roi.equal(feature_snapshot.roi)
        assert feature_latest.equal(feature_snapshot)

    def test_using_snapshot_modify_rois_does_not_modify_rois_in_snapshots(self, storage_with_snapshot):
        storage_with_snapshot.set_snapshot("snapshot")
        roi_snapshot = storage_with_snapshot.list_rois()[0]
        storage_with_snapshot.set_snapshot()
        roi_latest = storage_with_snapshot.list_rois()[0]

        assert roi_latest.equal(roi_snapshot)
        roi_latest.data.append(10.0)
        assert not roi_latest.equal(roi_snapshot)

    def test_using_snapshot_modify_features_does_not_modify_features_in_snapshots(self, storage_with_snapshot):
        storage_with_snapshot.set_snapshot("snapshot")
        feature_snapshot = storage_with_snapshot.list_features()[0]
        storage_with_snapshot.set_snapshot()
        feature_latest = storage_with_snapshot.list_features()[0]

        assert feature_snapshot.id != feature_latest.id
        assert feature_snapshot.roi.id != feature_latest.roi.id
        assert feature_latest.roi.equal(feature_snapshot.roi)
        assert feature_latest.equal(feature_snapshot)
        feature_latest.data_mz = 1000
        assert not feature_latest.equal(feature_snapshot)

    def test_from_sample_creates_same_sample_data(self, storage_with_snapshot):
        copy = memory.OnMemorySampleStorage.from_sample_storage(storage_with_snapshot)
        assert copy.list_rois() == storage_with_snapshot.list_rois()


class AssayStorageFixtures:
    @pytest.fixture
    def assay(self):
        return memory.OnMemoryAssayStorage("test-assay", helpers.ConcreteRoi, helpers.ConcreteFeature)

    @pytest.fixture(scope="class")
    def data_path(self, tmp_path_factory):
        return tmp_path_factory.mktemp("test-on-memory-sample-storage-snapshot")

    @pytest.fixture(scope="class")
    def sample_data1(self, data_path):
        sample = helpers.create_sample(data_path, 1)
        data = memory.OnMemorySampleStorage(sample, helpers.ConcreteRoi, helpers.ConcreteFeature)
        roi1 = helpers.create_roi(sample)
        roi2 = helpers.create_roi(sample)
        data.add_rois(roi1, roi2)
        ft1 = helpers.create_feature(roi1)
        ft2 = helpers.create_feature(roi2)
        ft1.annotation.group = 0  # type: ignore
        ft2.annotation.group = 0  # type: ignore
        ft3 = helpers.create_feature(roi2)
        data.add_features(ft1, ft2, ft3)
        return data

    @pytest.fixture(scope="class")
    def sample_data2(self, data_path):
        sample = helpers.create_sample(data_path, 2)
        data = memory.OnMemorySampleStorage(sample, helpers.ConcreteRoi, helpers.ConcreteFeature)
        roi1 = helpers.create_roi(sample)
        roi2 = helpers.create_roi(sample)
        data.add_rois(roi1, roi2)
        ft1 = helpers.create_feature(roi1)
        ft2 = helpers.create_feature(roi2)
        ft1.annotation.group = 0  # type: ignore
        ft2.annotation.group = 0  # type: ignore
        ft3 = helpers.create_feature(roi2)
        data.add_features(ft1, ft2, ft3)
        return data

    @pytest.fixture
    def assay_with_sample_data(self, assay, sample_data1):
        assay.add_sample_data(sample_data1)
        return assay

    @pytest.fixture
    def assay_with_two_samples(self, assay, sample_data1, sample_data2):
        assay.add_sample_data(sample_data1)
        assay.add_sample_data(sample_data2)
        return assay


class TestAssayStorageSampleData(AssayStorageFixtures):
    def test_create_ok(self):
        memory.OnMemoryAssayStorage("test-assay", helpers.ConcreteRoi, helpers.ConcreteFeature)

    def test_has_sample_no_sample_returns_false(self, assay):
        assert not assay.has_sample("invalid-sample-id")

    def test_add_sample_data_ok(self, assay, sample_data1):
        assay.add_sample_data(sample_data1)
        assert assay.has_sample(sample_data1.get_sample().id)
        for roi in sample_data1.list_rois():
            assert assay.has_roi(roi.id)

    def test_add_sample_with_same_id_raises_error(self, assay, sample_data1):
        assay.add_sample_data(sample_data1)
        with pytest.raises(exceptions.RepeatedIdError):
            assay.add_sample_data(sample_data1)

    def test_has_sample(self, assay_with_sample_data, sample_data1):
        assert assay_with_sample_data.has_sample(sample_data1.get_sample().id)

    def test_fetch_sample(self, assay_with_sample_data, sample_data1):
        expected = sample_data1.get_sample()
        actual = assay_with_sample_data.fetch_sample(expected.id)
        assert actual == expected

    def test_fetch_sample_no_sample_raises_error(self, assay):
        with pytest.raises(exceptions.SampleNotFound):
            assay.fetch_sample("invalid-sample-id")

    def test_rois_from_sample_data_has_equal_id_in_assay_storage(self, assay_with_sample_data, sample_data1):
        for roi in sample_data1.list_rois():
            assert assay_with_sample_data.has_roi(roi.id)

    def test_fetch_rois_by_sample(self, assay_with_sample_data, sample_data1):
        expected = sample_data1.list_rois()
        actual = assay_with_sample_data.fetch_rois_by_sample(sample_data1.get_sample().id)
        assert actual
        assert all(x.equal(y) for x, y in zip(expected, actual))

    def test_fetch_rois_by_sample_invalid_sample_raises_error(self, assay_with_sample_data):
        with pytest.raises(exceptions.SampleNotFound):
            assay_with_sample_data.fetch_rois_by_sample("invalid-sample-id")

    def test_fetch_rois_by_sample_modifying_rois_do_not_modify_rois_in_storage(
        self, assay_with_sample_data, sample_data1
    ):
        modified = assay_with_sample_data.fetch_rois_by_sample(sample_data1.get_sample().id)[0]
        modified.data.append(1000.0)
        original = assay_with_sample_data.fetch_rois_by_id(modified.id)[0]
        assert not modified.equal(original)

    def test_fetch_rois_by_id_modifying_roi_do_not_modify_roi_in_storage(self, assay_with_sample_data, sample_data1):
        roi_id = sample_data1.list_rois()[0].id
        modified = assay_with_sample_data.fetch_rois_by_id(roi_id)[0]
        modified.data.append(1000.0)
        original = assay_with_sample_data.fetch_rois_by_id(roi_id)[0]
        assert not modified.equal(original)

    def test_fetch_roi_by_id_invalid_id_raises_error(self, assay):
        with pytest.raises(exceptions.RoiNotFound):
            assay.fetch_rois_by_id(create_id())

    def test_list_samples(self, assay_with_sample_data, sample_data1):
        expected = [sample_data1.get_sample()]
        actual = assay_with_sample_data.list_samples()
        assert actual == expected

    def test_fetch_features_by_id(self, assay_with_sample_data, sample_data1):
        expected = sample_data1.list_features()
        actual = assay_with_sample_data.fetch_features_by_id(*(x.id for x in expected))
        assert all(x.equal(y) for x, y in zip(actual, expected))

    def test_fetch_features_by_id_invalid_feature_id_raises_error(self, assay):
        with pytest.raises(exceptions.FeatureNotFound):
            assay.fetch_features_by_id(create_id())

    def test_fetch_features_by_sample_with_shared_roi_share_the_same_roi_instance(
        self, assay_with_sample_data, sample_data1
    ):
        _, ft2, ft3 = assay_with_sample_data.fetch_features_by_sample(sample_data1.get_sample().id)
        assert ft2.roi == ft3.roi
        assert ft2.roi is ft3.roi

    def test_fetch_features_by_group(self, assay_with_sample_data):
        group = 0
        actual = assay_with_sample_data.fetch_features_by_group(group)
        assert all(ft.annotation.group == group for ft in actual)

    def test_fetch_features_by_group_invalid_group_raises_error(self, assay_with_sample_data):
        with pytest.raises(exceptions.FeatureGroupNotFound):
            invalid_group = 1000
            assay_with_sample_data.fetch_features_by_group(invalid_group)


class TestOnMemoryAssayStorageDescriptorApi(AssayStorageFixtures):
    def test_fetch_feature_descriptors(self, assay_with_two_samples):
        descriptors = assay_with_two_samples.fetch_descriptors()
        descriptor_names = helpers.ConcreteFeature.descriptor_names()
        n_features = assay_with_two_samples.get_n_features()
        for name, values in descriptors.items():
            assert name in descriptor_names
            assert len(values) == n_features

    def test_fetch_features_descriptors_from_single_sample(self, assay_with_two_samples, sample_data1):
        sample_id = sample_data1.get_sample().id
        descriptors = assay_with_two_samples.fetch_descriptors(sample_id=sample_id)
        descriptor_names = helpers.ConcreteFeature.descriptor_names()
        n_features = sample_data1.get_n_features()
        for name, values in descriptors.items():
            assert name in descriptor_names
            assert len(values) == n_features

    def test_fetch_descriptors_subset(self, assay_with_two_samples):
        subset = {"mz", "height"}
        descriptors = assay_with_two_samples.fetch_descriptors(descriptors=subset)

        n_features = assay_with_two_samples.get_n_features()
        excluded_descriptor = "custom_descriptor"

        assert excluded_descriptor in helpers.ConcreteFeature.descriptor_names()
        assert excluded_descriptor not in descriptors

        for name, values in descriptors.items():
            assert name in subset
            assert len(values) == n_features

    def test_fetch_descriptors_subset_from_single_sample(self, assay_with_two_samples, sample_data1):
        subset = {"mz", "height"}
        sample_id = sample_data1.get_sample().id
        descriptors = assay_with_two_samples.fetch_descriptors(sample_id=sample_id, descriptors=subset)

        n_features = sample_data1.get_n_features()
        excluded_descriptor = "custom_descriptor"

        assert excluded_descriptor in helpers.ConcreteFeature.descriptor_names()
        assert excluded_descriptor not in descriptors

        for name, values in descriptors.items():
            assert name in subset
            assert len(values) == n_features

    def test_fetch_descriptors_invalid_sample_raises_error(self, assay_with_sample_data):
        with pytest.raises(exceptions.SampleNotFound):
            assay_with_sample_data.fetch_descriptors(sample_id="invalid-sample-id")

    def test_fetch_descriptors_invalid_descriptor_name_raises_error(self, assay_with_sample_data):
        with pytest.raises(exceptions.InvalidFeatureDescriptor):
            assay_with_sample_data.fetch_descriptors(descriptors=["invalid-descriptor"])

    def test_fetch_descriptors_retrieves_a_copy(self, assay_with_sample_data):
        descriptors = assay_with_sample_data.fetch_descriptors()

        # modify a random value in the descriptors table and check if values in the assay are modified
        name = list(descriptors)[0]
        original = descriptors[name][0]
        modified = 100000.0
        descriptors[name][0] = modified

        descriptors2 = assay_with_sample_data.fetch_descriptors()
        assert descriptors is not descriptors2
        assert descriptors != descriptors2
        assert descriptors2[name][0] == original

    def test_patch_descriptors_update_data(self, assay_with_sample_data):
        annotations = assay_with_sample_data.fetch_annotations()
        descriptors_before = assay_with_sample_data.fetch_descriptors()
        descriptor_name = "mz"  # arbitrary descriptor and index
        index = 2
        id_ = annotations[index].id

        original = descriptors_before[descriptor_name][index]
        modified = 100000.0
        patch = DescriptorPatch(id=id_, descriptor=descriptor_name, value=modified)

        assay_with_sample_data.patch_descriptors(patch)

        descriptors_after = assay_with_sample_data.fetch_descriptors()

        assert original != modified
        assert descriptors_before[descriptor_name][index] == original
        assert descriptors_after[descriptor_name][index] == modified

    def test_patch_descriptor_and_revert_to_previous_snapshot_retrieves_original_data(self, assay_with_sample_data):
        # first we create an snapshot
        snapshot_id = "snapshot"
        assay_with_sample_data.create_snapshot(snapshot_id)

        # then we apply a patch
        annotations = assay_with_sample_data.fetch_annotations()
        descriptors_before_patch = assay_with_sample_data.fetch_descriptors()
        index = 2
        id_ = annotations[index].id

        patch = DescriptorPatch(id=id_, descriptor="mz", value=100000.0)

        assay_with_sample_data.patch_descriptors(patch)
        descriptors_after_patch = assay_with_sample_data.fetch_descriptors()

        assay_with_sample_data.set_snapshot(snapshot_id)
        descriptors_snapshot = assay_with_sample_data.fetch_descriptors()

        assert descriptors_before_patch == descriptors_snapshot
        assert descriptors_before_patch != descriptors_after_patch

    def test_patch_descriptor_on_non_latest_snapshot_raises_error(self, assay_with_sample_data):
        # first we create an snapshot
        snapshot_id = "snapshot"
        assay_with_sample_data.create_snapshot(snapshot_id)
        assay_with_sample_data.set_snapshot(snapshot_id)

        # then we apply a patch
        annotations = assay_with_sample_data.fetch_annotations()
        index = 2
        id_ = annotations[index].id

        patch = DescriptorPatch(id=id_, descriptor="mz", value=100000.0)

        with pytest.raises(exceptions.SnapshotError):
            assay_with_sample_data.patch_descriptors(patch)

    def test_patch_descriptor_invalid_id_raises_error(self, assay_with_sample_data):
        invalid_id = create_id()
        patch = DescriptorPatch(id=invalid_id, descriptor="mz", value=100000.0)
        with pytest.raises(exceptions.FeatureNotFound):
            assay_with_sample_data.patch_descriptors(patch)


class TestOnMemoryAssayStorageAnnotationApi(AssayStorageFixtures):
    def test_fetch_annotations(self, assay_with_two_samples):
        actual = assay_with_two_samples.fetch_annotations()
        assert len(actual) == assay_with_two_samples.get_n_features()

    def test_fetch_annotations_from_single_sample(self, assay_with_two_samples, sample_data1):
        sample_id = sample_data1.get_sample().id
        actual = assay_with_two_samples.fetch_annotations(sample_id=sample_id)
        assert len(actual) == sample_data1.get_n_features()

    def test_fetch_annotations_from_invalid_sample_raises_error(self, assay_with_two_samples):
        with pytest.raises(exceptions.SampleNotFound):
            assay_with_two_samples.fetch_annotations(sample_id="invalid-sample-id")

    def test_fetch_annotations_retrieves_a_copy(self, assay_with_sample_data):
        annotations_before = assay_with_sample_data.fetch_annotations()
        index = 1
        annotations_before[index].group = 10000
        annotations_after = assay_with_sample_data.fetch_annotations()

        assert annotations_before != annotations_after

    def test_patch_annotations_update_data(self, assay_with_sample_data):
        annotations_before_patch = assay_with_sample_data.fetch_annotations()
        index = 1
        id_ = annotations_before_patch[index].id
        field = "group"
        value = 1000
        patch = AnnotationPatch(id=id_, field=field, value=value)
        assay_with_sample_data.patch_annotations(patch)
        annotations_after_patch = assay_with_sample_data.fetch_annotations()
        assert annotations_before_patch != annotations_after_patch
        assert annotations_after_patch[index].group == value

    def test_patch_annotations_and_revert_to_previous_snapshot_retrieves_original_data(self, assay_with_sample_data):
        snapshot_id = "snapshot"
        assay_with_sample_data.create_snapshot(snapshot_id)
        annotations_before_patch = assay_with_sample_data.fetch_annotations()
        index = 1
        id_ = annotations_before_patch[index].id
        field = "group"
        value = 1000
        patch = AnnotationPatch(id=id_, field=field, value=value)
        assay_with_sample_data.patch_annotations(patch)
        annotations_after_patch = assay_with_sample_data.fetch_annotations()

        assay_with_sample_data.set_snapshot(snapshot_id)
        annotations_snapshot = assay_with_sample_data.fetch_annotations()

        assert annotations_before_patch != annotations_after_patch
        assert annotations_before_patch == annotations_snapshot

    def test_patch_annotation_on_non_latest_snapshot_raises_error(self, assay_with_sample_data):
        snapshot_id = "snapshot"
        assay_with_sample_data.create_snapshot(snapshot_id)
        assay_with_sample_data.set_snapshot(snapshot_id)
        annotations = assay_with_sample_data.fetch_annotations()
        index = 1
        id_ = annotations[index].id
        field = "group"
        value = 1000
        patch = AnnotationPatch(id=id_, field=field, value=value)
        with pytest.raises(exceptions.SnapshotError):
            assay_with_sample_data.patch_annotations(patch)

    def test_patch_annotation_invalid_id_raises_error(self, assay_with_sample_data):
        invalid_id = create_id()
        patch = AnnotationPatch(id=invalid_id, field="group", value=10000)
        with pytest.raises(exceptions.FeatureNotFound):
            assay_with_sample_data.patch_annotations(patch)


class TestOnMemoryAssayStorageFillValueApi(AssayStorageFixtures):
    def test_fetch_fill_values_no_values_return_empty_dict(self, assay):
        fill_values = assay.fetch_fill_values()
        assert isinstance(fill_values, dict)
        assert not fill_values

    def test_fetch_add_values_and_fetch_return_same_fill_values(self, assay):
        val1 = FillValue(sample_id="sample1", feature_group=0, value=10.0)
        val2 = FillValue(sample_id="sample2", feature_group=1, value=5.0)
        assay.add_fill_values(val1, val2)
        actual = assay.fetch_fill_values()
        assert actual[val1.sample_id][val1.feature_group] == val1.value
        assert actual[val2.sample_id][val2.feature_group] == val2.value

    def test_fetch_fill_values_return_a_copy(self, assay):
        original = 10.0
        modified = 100.0
        val1 = FillValue(sample_id="sample1", feature_group=0, value=original)
        assay.add_fill_values(val1)
        fill_values_copy1 = assay.fetch_fill_values()
        fill_values_copy1[val1.sample_id][val1.feature_group] = modified
        fill_values_copy2 = assay.fetch_fill_values()
        assert fill_values_copy2[val1.sample_id][val1.feature_group] == original
        assert fill_values_copy1 is not fill_values_copy2

    def test_snapshots_maintain_independent_fill_values_copies(self, assay):
        # add fill values and create a snapshot
        val1 = FillValue(sample_id="sample1", feature_group=0, value=10.0)
        assay.add_fill_values(val1)
        snapshot_id = "snapshot"
        assay.create_snapshot(snapshot_id)

        # add more fill values and set previous snapshot
        val2 = FillValue(sample_id="sample2", feature_group=1, value=5.0)
        assay.add_fill_values(val2)
        assay.set_snapshot(snapshot_id)

        snapshot_fill_values = assay.fetch_fill_values()
        assert snapshot_fill_values[val1.sample_id][val1.feature_group] == val1.value
        assert val2.sample_id not in snapshot_fill_values

        # go back to latest status and check that value 2 is stored
        assay.set_snapshot()
        latest_fill_values = assay.fetch_fill_values()
        assert latest_fill_values[val1.sample_id][val1.feature_group] == val1.value
        assert latest_fill_values[val2.sample_id][val2.feature_group] == val2.value


class TestOnMemoryAssayStorageMiscApi(AssayStorageFixtures):
    def test_get_n_rois(self, assay_with_sample_data):
        expected = 2
        actual = assay_with_sample_data.get_n_rois()
        assert actual == expected

    def test_get_n_rois_empty_assay_returns_zero(self, assay):
        assert not assay.get_n_rois()

    def test_create_snapshot_using_reserved_id_raises_error(self, assay_with_sample_data):
        with pytest.raises(ValueError):
            assay_with_sample_data.create_snapshot(memory.LATEST)

    def test_create_snapshot_with_existing_id_raises_error(self, assay_with_sample_data):
        snapshot_id = "snapshot"
        assay_with_sample_data.create_snapshot(snapshot_id)
        with pytest.raises(exceptions.RepeatedIdError):
            assay_with_sample_data.create_snapshot(snapshot_id)

    def test_set_snapshot_invalid_id_raises_error(self, assay_with_sample_data):
        with pytest.raises(exceptions.SnapshotNotFound):
            assay_with_sample_data.set_snapshot("invalid-id")

    def test_set_snapshot_reset_deletes_snapshots(self, assay_with_sample_data):
        assay_with_sample_data.create_snapshot("snapshot1")
        assay_with_sample_data.create_snapshot("snapshot2")
        assay_with_sample_data.set_snapshot("snapshot1", reset=True)
        assert "snapshot2" not in assay_with_sample_data.list_snapshots()

    def test_set_reset_no_arguments_set_to_latest_snapshot(self, assay_with_sample_data):
        snapshot_id = "snapshot"
        assay_with_sample_data.create_snapshot(snapshot_id)
        assay_with_sample_data.set_snapshot(snapshot_id)
        assert assay_with_sample_data.get_snapshot_id() == snapshot_id
        assay_with_sample_data.set_snapshot()
        assert assay_with_sample_data.get_snapshot_id() == memory.LATEST



"""TidyMS core operators."""

from __future__ import annotations

from typing import Any, overload

from ..dataflow import AssayProcessStatus, DataMatrixProcessStatus, SampleProcessStatus
from ..exceptions import PipelineConfigurationError, ProcessStatusError, RepeatedIdError
from ..matrix import DataMatrix
from ..models import FeatureType, RoiType
from ..registry import operator_registry
from ..storage import AssayStorage, SampleStorage
from .assay import AssayOperator
from .base import BaseOperator
from .matrix import MatrixOperator
from .sample import SampleOperator


class Pipeline:
    """Compose multiple operators of the same type into a single unit."""

    def __init__(self, id: str) -> None:
        self.id = id
        self.operators: list[SampleOperator | AssayOperator | MatrixOperator | Pipeline] = list()
        self._id_to_operator = dict()

    def copy(self) -> Pipeline:
        """Create a copy of the current instance."""
        cp = Pipeline(self.id)
        for op in self.operators:
            op_cp = op.copy() if isinstance(op, Pipeline) else op.model_copy(deep=True)
            cp.add_operator(op_cp)
        return cp

    def __eq__(self, other) -> bool:
        equal_ids = self.id == other.id
        equal_operators = self.operators == other.operators
        return equal_ids and equal_operators

    @overload
    def add_operator(self, operator: MatrixOperator) -> None: ...

    @overload
    def add_operator(self, operator: AssayOperator) -> None: ...

    @overload
    def add_operator(self, operator: SampleOperator) -> None: ...

    @overload
    def add_operator(self, operator: Pipeline) -> None: ...

    def add_operator(self, operator):
        """Add a new operator to the pipeline.

        :param operator: the operator to add
        :raises PipelineConfigurationError: if the operator type differs from the pipeline type.

        """
        check_compatible_element(self, operator)

        if operator.id in self._id_to_operator:
            msg = f"Pipeline {self.id} already contains an operator with id {operator.id}."
            raise RepeatedIdError(msg)

        self.operators.append(operator)
        self._id_to_operator[operator.id] = operator

    def get_operator(self, id: str) -> AssayOperator | SampleOperator | Pipeline:
        """Get an operator by id.

        :param id: the id of an operator in the pipeline
        :raises ValueError: if the operator is not found

        """
        try:
            return self._id_to_operator[id]
        except KeyError:
            raise ValueError(f"Operator with id {id} not found in pipeline `{self.id}`.")

    def list_operator_ids(self) -> list[str]:
        """Retrieve the name of all operators in the pipeline."""
        return list(self._id_to_operator)

    @overload
    def apply(self, data: SampleStorage[RoiType, FeatureType]) -> None: ...

    @overload
    def apply(self, data: AssayStorage[RoiType, FeatureType]) -> None: ...

    @overload
    def apply(self, data: DataMatrix) -> None: ...

    def apply(self, data) -> None:
        """Apply pipeline to the data."""
        for op in self.operators:
            op.apply(data)

    @classmethod
    def deserialize(cls, d: dict[str, Any]) -> Pipeline:
        """Deserialize a dictionary into a pipeline."""
        id_ = d.get("id")
        if not isinstance(id_, str):
            raise ValueError("`id` is a mandatory field and must be a string.")

        operators = d.get("operators")
        if not isinstance(operators, list):
            raise ValueError("`operators` is a mandatory field and must be a list of dictionaries.")

        pipe = Pipeline(id_)

        for d in operators:
            if not isinstance(d, dict):
                raise ValueError("`operators` element is not a dictionary.")
            op_type = d.pop("class", None)
            if op_type is None:
                op = Pipeline.deserialize(d)
            else:
                T = operator_registry.get(op_type)
                op = T(**d)
            pipe.add_operator(op)
        return pipe

    def serialize(self) -> dict:
        """Serialize pipeline into a JSON serializable dictionary."""
        operators = list()
        serialized = {"id": self.id, "operators": operators}
        for op in self.operators:
            if isinstance(op, Pipeline):
                d = op.serialize()
            else:
                d = op.model_dump(mode="json")
                d["class"] = op.__class__.__name__
            operators.append(d)
        return serialized

    def validate_dataflow(self) -> None:
        """Check if the data status throughout the pipeline is valid."""
        first_operator = self.operators[0]
        while not isinstance(first_operator, BaseOperator):
            first_operator = first_operator.operators[0]

        initial_status = get_initial_process_status(first_operator)
        try:
            self._validate_dataflow_recursion(initial_status)
        except ProcessStatusError as e:
            msg = "Check that all operators are of the same type and that the order is valid."
            raise PipelineConfigurationError(msg) from e

    def _validate_dataflow_recursion(self, status):
        for op in self.operators:
            if isinstance(op, Pipeline):
                op._validate_dataflow_recursion(status)
            else:
                op.check_status(status)
                op.update_status(status)


@overload
def get_initial_process_status(operator: SampleOperator) -> SampleProcessStatus: ...


@overload
def get_initial_process_status(operator: AssayOperator) -> AssayProcessStatus: ...


@overload
def get_initial_process_status(operator: MatrixOperator) -> AssayProcessStatus: ...


def get_initial_process_status(operator):
    """Create an initial data process status."""
    if isinstance(operator, SampleOperator):
        return SampleProcessStatus()
    elif isinstance(operator, AssayOperator):
        return AssayProcessStatus()
    elif isinstance(operator, MatrixOperator):
        return DataMatrixProcessStatus()
    else:
        raise NotImplementedError


OT = SampleOperator | AssayOperator | MatrixOperator
PipelineElement = OT | Pipeline


def check_compatible_element(pipeline: Pipeline, element: PipelineElement) -> None:
    """Raise an exception if two pipeline elements are not compatible."""
    pipeline_first = get_first_operator(pipeline)
    element_first = get_first_operator(element)
    if element_first is None:
        raise PipelineConfigurationError("Nested pipelines cannot be empty")

    if pipeline_first is None:
        return None

    both_sample_operators = isinstance(pipeline_first, SampleOperator) and isinstance(element_first, SampleOperator)
    both_assay_operators = isinstance(pipeline_first, AssayOperator) and isinstance(element_first, AssayOperator)
    both_matrix_operators = isinstance(pipeline_first, MatrixOperator) and isinstance(element_first, MatrixOperator)
    if not (both_sample_operators or both_assay_operators or both_matrix_operators):
        msg = "All pipeline elements must process the same data type (sample, assay or matrix)."
        raise PipelineConfigurationError(msg)


def get_first_operator(op: PipelineElement) -> OT | None:
    """Get the first element of a pipeline or raise an error if the pipeline is empty."""
    if not isinstance(op, Pipeline):
        return op
    elif not op.operators:
        return None
    return get_first_operator(op.operators[0])



"""Processor class for sample isotopologue annotation."""

from typing import TYPE_CHECKING, Self

from ..core.dataflow import SampleProcessStatus
from ..core.enums import MSInstrument, Polarity, SeparationMode
from ..core.models import AnnotableFeatureType, RoiType
from ..core.operators.sample import SampleOperator
from ..core.storage import SampleStorage
from .annotation import AnnotatorParameters, annotate, create_annotation_tools

if TYPE_CHECKING:
    from typing import assert_never


class IsotopologueAnnotator(SampleOperator[RoiType, AnnotableFeatureType], AnnotatorParameters):
    """Annotate isotopologues in a sample.

    Groups isotopologue features. Each group is assigned an unique label, and a
    charge state. Each feature in a group is assigned an unique index that
    determines the position in the envelope.

    Annotations are stored to the `annotation` attribute of each feature.
    """

    def _apply_operator(self, data: SampleStorage[RoiType, AnnotableFeatureType]):
        tools = create_annotation_tools(self)
        annotate(data.list_features(), *tools)

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        """Set the annotator default parameters.

        :param instrument : the instrument type used in the experimental setup
        :param separation : the LC platform used in the experimental setup
        :param polarity : the MS polarity used in the experiment

        """
        op = cls.from_chnops(2000)
        op.max_length = 10

        match polarity:
            case Polarity.POSITIVE:
                op.max_charge = 3
            case Polarity.NEGATIVE:
                op.max_charge = -3
            case _ as never:
                assert_never(never)

        match instrument:
            case MSInstrument.QTOF:
                op.min_M_tol = 0.005
                op.max_M_tol = 0.01
            case MSInstrument.ORBITRAP:
                op.min_M_tol = 0.001
                op.max_M_tol = 0.005
            case _ as never:
                assert_never(never)

        return op

    def get_expected_status_in(self) -> SampleProcessStatus:
        """Get the expected status before isotopologue annotation."""
        return SampleProcessStatus(roi_extracted=True, feature_extracted=True)

    def get_expected_status_out(self) -> SampleProcessStatus:
        """Get the expected status after isotopologue annotation."""
        return SampleProcessStatus(roi_extracted=True, feature_extracted=True, isotopologue_annotated=True)



"""General purpose raw data reader."""

from __future__ import annotations

import pathlib
from collections import OrderedDict
from contextlib import contextmanager
from typing import Generator

from typing_extensions import Callable

from ..core.enums import MSDataMode
from ..core.models import Chromatogram, MSSpectrum, Sample
from .reader import Reader, reader_registry


class MSData:
    """Provide access to raw MS data.

    Data is read from disk in a lazy manner and cached in memory.

    :param src: raw data source. It may be the path to a raw data file or a sample model.
        If the latter is provided, the path to the data source is fetch from the
        :py:attr:`tidyms2.core.models.Sample.path` field.
    :param reader: the Reader to read raw data. If ``None``, the reader is inferred using the file extension.
        If `src` is a sample model and :py:attr:`~tidyms2.core.models.Sample.reader` is defined, the reader will
        be fetch fetched from the reader registry.
    :param mode: the mode in which the data is stored. If `src` is a sample instance, this parameter is ignored
        and is fetched from the sample data.
    :param centroider: a function that takes a spectrum in profile mode and converts it to centroid mode. Only
        used if :py:attr:`mode` is set to profile mode.
    :param cache: int, default=-1
        The maximum cache size, in bytes. The cache will store spectrum data until it surpasses this value. At this
        point, old entries will be deleted from the cache. If set to``-1``, the cache can grow indefinitely.
    :param ms_level: skip spectra without this MS level when iterating over spectra.  If `src` is a sample instance,
        this parameter is ignored and is fetched from the sample data.
    :param start_time: skip spectra with time lower than this value when iterating over data.  If `src` is a sample
        instance, this parameter is ignored and is fetched from the sample data.
    :param end_time: skip spectra with time greater than this value when iterating over data.  If `src` is a sample
        instance, this parameter is ignored and is fetched from the sample data.
    :param kwargs: keyword arguments passed to the reader.

    """

    def __init__(
        self,
        src: pathlib.Path | Sample,
        reader: type[Reader] | None = None,
        mode: MSDataMode = MSDataMode.CENTROID,
        centroider: Callable[[MSSpectrum], MSSpectrum] | None = None,
        cache: int = -1,
        ms_level: int = 1,
        start_time: float = 0.0,
        end_time: float | None = None,
        **kwargs,
    ):
        if isinstance(src, pathlib.Path):
            id_ = src.name
            src = Sample(
                id=id_, path=src, ms_level=ms_level, start_time=start_time, end_time=end_time, ms_data_mode=mode
            )
        self.sample = src

        if reader is None and isinstance(self.sample.reader, str):
            reader = reader_registry.get(self.sample.reader)
        elif reader is None:
            reader = reader_registry.get(self.sample.path.suffix[1:])

        self.centroider = centroider
        self._reader = reader(self.sample, **kwargs)
        self._cache = MSDataCache(max_size=cache)
        self._n_spectra: int | None = None
        self._n_chromatogram: int | None = None

    def get_sample(self) -> Sample:
        """Retrieve the sample associated with the data."""
        return self.sample

    def get_n_chromatograms(self) -> int:
        """Retrieve the total number of chromatograms stored in the source."""
        if self._n_chromatogram is None:
            self._n_chromatogram = self._reader.get_n_chromatograms()
        return self._n_chromatogram

    def get_n_spectra(self) -> int:
        """Retrieve the total number of spectra stored in the source."""
        if self._n_spectra is None:
            self._n_spectra = self._reader.get_n_spectra()
        return self._n_spectra

    def get_chromatogram(self, index: int) -> Chromatogram:
        """Retrieve a chromatogram by index."""
        return self._reader.get_chromatogram(index)

    def get_spectrum(self, index: int) -> MSSpectrum:
        """Retrieve a spectrum by index."""
        n_sp = self.get_n_spectra()
        if (index < 0) or (index >= n_sp):
            msg = f"`index` must be integer in the interval [0:{n_sp}). Got {index}."
            raise ValueError(msg)

        if self._cache.check(index):
            spectrum = self._cache.get(index)
        else:
            spectrum = self._reader.get_spectrum(index)
            if self.sample.ms_data_mode is MSDataMode.PROFILE and self.centroider is not None:
                spectrum = self.centroider(spectrum)
            spectrum.centroid = self.sample.ms_data_mode is MSDataMode.CENTROID
            self._cache.add(spectrum)
        return spectrum

    def __iter__(self) -> Generator[MSSpectrum, None, None]:
        """Iterate over all spectra in the data."""
        for k in range(self.get_n_spectra()):
            sp = self.get_spectrum(k)
            if (self.sample.ms_level == sp.ms_level) and (self.sample.start_time <= sp.time):
                if (self.sample.end_time is None) or (self.sample.end_time > sp.time):
                    yield sp

    @contextmanager
    def using_tmp_config(
        self, ms_level: int | None = None, start_time: float | None = None, end_time: float | None = None
    ):
        """Context manager that temporarily modifies MS level and scans time range.

        :param ms_level: temporary value for the MS level. If set to ``None`` the original value is not modified.
        :param start_time: temporary value for the start time. If set to ``None`` the original value is not modified.
        :param end_time: temporary value for the end time. If set to ``None`` the original value is not modified.

        """
        sample = self.get_sample()
        tmp_config = {
            "ms_level": ms_level or sample.ms_level,
            "start_time": start_time or sample.start_time,
            "end_time": end_time or sample.end_time,
        }
        sample_with_tmp_config = sample.model_copy(update=tmp_config)
        Sample.model_validate(sample_with_tmp_config)
        try:
            self.sample = sample_with_tmp_config
            yield
        finally:
            self.sample = sample


class MSDataCache:
    """Cache spectra data to avoid reading from disk.

    Old entries are deleted if the cache grows larger than total data size in bytes. The maximum size of the cache is
    defined by `max_size`. If set to ``-1``, the cache can grow indefinitely.

    """

    def __init__(self, max_size: int = -1):
        self.cache: OrderedDict[int, MSSpectrum] = OrderedDict()
        self.size = 0
        self.max_size = max_size

    def add(self, spectrum: MSSpectrum) -> None:
        """Store a spectrum."""
        self.cache[spectrum.index] = spectrum
        self.size += spectrum.get_nbytes()
        self._prune()

    def get(self, index: int) -> MSSpectrum:
        """Retrieve a spectrum from the cache. If not found, returns ``None``."""
        spectrum = self.cache[index]
        self.cache.move_to_end(index)
        return spectrum

    def check(self, index: int) -> bool:
        """Check if the provided index is in the cache."""
        return index in self.cache

    def _prune(self) -> None:
        """Delete entries until the cache size is lower than max_size."""
        if self.max_size > -1:
            while self.size > self.max_size:
                _, spectrum = self.cache.popitem(last=False)
                self.size -= spectrum.get_nbytes()



import pytest

from tidyms2.core.models import Sample
from tidyms2.io.msdata import MSData, MSDataCache
from tidyms2.io.mzml import MZMLReader


class TestMSDataFromSample:
    @pytest.fixture
    def sample(self, raw_data_dir):
        sample_path = raw_data_dir / "centroid-data-zlib-indexed-compressed.mzML"
        return Sample(path=sample_path, id="example")

    def test_create_from_sample(self, sample):
        ms_data = MSData(sample)
        n_sp = ms_data.get_n_spectra()
        assert n_sp > 0

    def test_with_start_time(self, sample: Sample):
        expected = 10.0
        sample = sample.model_copy(update={"start_time": expected})
        ms_data = MSData(sample)
        for sp in ms_data:
            assert sp.time >= expected

    def test_with_end_time(self, sample: Sample):
        expected = 10.0
        sample = sample.model_copy(update={"end_time": expected})
        ms_data = MSData(sample)
        for sp in ms_data:
            assert sp.time <= expected

    def test_with_ms_level(self, sample: Sample):
        sample = sample.model_copy(update={"ms_level": 1})
        ms_data = MSData(sample)
        spectra = [sp for sp in ms_data]
        assert spectra

    def test_iterate_ignore_all_spectra_if_no_spectra_with_ms_level(self, sample: Sample):
        sample = sample.model_copy(update={"ms_level": 10})
        ms_data = MSData(sample)
        spectra = [sp for sp in ms_data]
        assert not spectra


class TestMSDataCache:
    @pytest.fixture
    def sample_path(self, raw_data_dir):
        return raw_data_dir / "profile-data-zlib-indexed-compressed.mzML"

    def test_cache_add_data_no_limits(self, sample_path):
        reader = MZMLReader(sample_path)
        cache = MSDataCache()

        index = 5
        spectrum = reader.get_spectrum(index)
        cache.add(spectrum)
        assert cache.size == spectrum.get_nbytes()

        stored = cache.get(index)
        assert spectrum == stored

    def test_cache_with_limit_delete_entries(self, sample_path):
        reader = MZMLReader(sample_path)

        index1 = 0
        index2 = 1
        sp1 = reader.get_spectrum(index1)
        sp2 = reader.get_spectrum(index2)
        size1 = sp1.get_nbytes()
        size2 = sp2.get_nbytes()
        max_size = max(size1, size2)
        cache = MSDataCache(max_size=max_size)
        cache.add(sp1)
        cache.add(sp2)

        assert cache.check(index2)
        assert not cache.check(index1)



"""Utilities to process LC-MS datasets."""

from .assay import create_lcms_assay
from .models import Peak
from .operators import LCFeatureMatcher, LCPeakExtractor, LCTraceBaselineEstimator, LCTraceExtractor, LCTraceSmoother

__all__ = [
    "create_lcms_assay",
    "LCFeatureMatcher",
    "LCTraceBaselineEstimator",
    "LCTraceExtractor",
    "LCTraceSmoother",
    "Peak",
    "LCPeakExtractor",
]



import numpy as np
import matplotlib.pyplot as plt

# always generate the same plot
np.random.seed(1234)
time = np.arange(50)

amp = 30.0
width = 2.0
rt = 25.0
signal =  amp * np.power(np.e, -0.5 * ((time - rt) / width) ** 2)
noise = np.random.normal(size=signal.size, scale=1)
x = signal + noise + 3
mz_mean = 203.08215
mz = np.random.normal(size=signal.size, scale=0.0005) + mz_mean

fig, ax = plt.subplots(figsize=(6, 6), nrows=2, sharex=True)
ax[1].plot(time, x)
ax[1].set_ylabel("Intensity")
ax[1].set_xlabel("Retention Time")
ax[0].plot(time, mz)
ax[0].set_ylabel("m/z")
ax[0].set_ylim(mz_mean - 0.0025, mz_mean + 0.0025)



from tidyms2.io.matrix import read_progenesis


def test_read_progenesis(data_dir):
    path = data_dir / "matrix/progenesis.csv"
    matrix = read_progenesis(path)
    assert matrix.get_n_samples() == 50
    assert matrix.get_n_features() == 100
